Package: TopDom
Version: 0.10.2
Title: An Efficient and Deterministic Method for Identifying
        Topological Domains in Genomes
Depends: R (>= 3.3.0)
Imports: matrixStats, graphics, grid, ggplot2, reshape2, stats, tibble,
        utils
Suggests: diffobj (>= 0.1.11)
Authors@R: c(person("Henrik", "Bengtsson", role = c("aut", "cre", "cph"),
                    email = "henrikb@braju.com",
                    comment = c(ORCID = "0000-0002-7579-5165")),
             person("Hanjun", "Shin", role = c("aut", "ctr", "cph"),
                    email = "shanjun@usc.edu"),
             person("Harris", "Lazaris", role = c("ctr", "cph"),
                    comment = c("PhD Student, NYU",
                                ORCID = "0000-0002-7345-8288")),
             person("Gangqing", "Hu", role = c("ctr", "cph"),
                    comment = c("Staff Scientist, NIH",
                                ORCID = "0000-0001-5453-6888")),
             person("Xianghong", "Zhou", role = c("ctr"),
                    email = "XJZhou@mednet.ucla.edu"))
Description: The 'TopDom' method identifies topological domains in genomes from Hi-C sequence data (Shin et al., 2016 <doi:10.1093/nar/gkv1505>).  The authors published an implementation of their method as an R script (two different versions; also available in this package).  This package originates from those original 'TopDom' R scripts and provides help pages adopted from the original 'TopDom' PDF documentation.  It also provides a small number of bug fixes to the original code.
URL: https://github.com/HenrikBengtsson/TopDom
BugReports: https://github.com/HenrikBengtsson/TopDom/issues
License: GPL
Language: en-US
Encoding: UTF-8
LazyLoad: TRUE
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-08-30 17:48:56 UTC; hb
Author: Henrik Bengtsson [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-7579-5165>),
  Hanjun Shin [aut, ctr, cph],
  Harris Lazaris [ctr, cph] (PhD Student, NYU, ORCID:
    <https://orcid.org/0000-0002-7345-8288>),
  Gangqing Hu [ctr, cph] (Staff Scientist, NIH, ORCID:
    <https://orcid.org/0000-0001-5453-6888>),
  Xianghong Zhou [ctr]
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Repository: CRAN
Date/Publication: 2026-08-31 15:40:31 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",TopDom)
S3method("[",TopDomData)
S3method(as_tibble,TopDomOverlapScores)
S3method(countsPerRegion,TopDomData)
S3method(dim,TopDom)
S3method(dim,TopDomData)
S3method(ggCountHeatmap,TopDomData)
S3method(image,TopDomData)
S3method(print,TopDom)
S3method(print,TopDomData)
S3method(print,TopDomOverlapScores)
S3method(subsetByRegion,TopDom)
S3method(subsetByRegion,TopDomData)
export(TopDom)
export(countsPerRegion)
export(ggCountHeatmap)
export(ggDomain)
export(ggDomainLabel)
export(legacy)
export(overlapScores)
export(readHiC)
export(subsetByRegion)
importFrom(ggplot2,
  aes,
  annotation_custom,
  coord_fixed,
  geom_raster,
  geom_segment,
  ggplot,
  scale_fill_gradient2,
  theme_void
)
importFrom(graphics,image)
importFrom(grid,
  gpar,
  textGrob
)
importFrom(matrixStats,colSums2)
importFrom(reshape2,melt)
importFrom(stats,wilcox.test)
importFrom(tibble,
  as_tibble,
  tibble
)
importFrom(utils,
  capture.output,
  file_test,
  read.table,
  str,
  write.table
)

<div id="badges"><!-- pkgdown markup -->
<a href="https://CRAN.R-project.org/web/checks/check_results_TopDom.html"><img border="0" src="https://www.r-pkg.org/badges/version/TopDom" alt="CRAN check status"/></a> <a href="https://github.com/HenrikBengtsson/TopDom/actions?query=workflow%3AR-CMD-check"><img border="0" src="https://github.com/HenrikBengtsson/TopDom/actions/workflows/R-CMD-check.yaml/badge.svg?branch=develop" alt="R CMD check status"/></a>     <a href="https://app.codecov.io/gh/HenrikBengtsson/TopDom"><img border="0" src="https://codecov.io/gh/HenrikBengtsson/TopDom/branch/develop/graph/badge.svg" alt="Coverage Status"/></a> 
</div>

# TopDom: An Efficient and Deterministic Method for Identifying Topological Domains in Genomes 

The TopDom method identifies topological domains in genomes from Hi-C sequence data (Shin et al., 2016).  The authors published an implementation of their method as an R script (two different versions; also available in this package).  This package originates from those original TopDom R scripts and provides help pages adopted from the original TopDom PDF documentation.  It also provides a small number of bug fixes to the original code.


## Citing TopDom - the method and the package

Whenever using the TopDom method, please cite:

* Hanjun Shin, Yi Shi, Chao Dai, Harianto Tjong, Ke Gong, Frank Alber, Xianghong Jasmine Zhou; TopDom: an efficient and deterministic method for identifying topological domains in genomes, Nucleic Acids Research, Volume 44, Issue 7, 20 April 2016, Pages e70, [10.1093/nar/gkv1505](https://doi.org/10.1093/nar/gkv1505). PMCID: PMC4838359, PMID: 26704975.

Whenever using the **TopDom** package, please cite:

* Henrik Bengtsson, Hanjun Shin, Harris Lazaris, Gangqing Hu and Xianghong Zhou (2020). R Package TopDom: An Efficient and Deterministic Method for Identifying Topological Domains in Genomes. R package version 0.10.1. <https://CRAN.R-project.org/package=TopDom>

The above information is also available as plain text as well as BibTeX entries via `citation("TopDom")`.


## Background

Please note that I, Henrik Bengtsson, is _not_ the developer of the TopDom method (Shin et al. 2016), or the (original) TopDom code available from the TopDom authors.   As I needed the TopDom method in a project, I ended up putting their `TopDom_v0.0.*.R` scripts into a proper R package (this **TopDom** package) so I could validate the[ir] code via `R CMD check` and so on. Then I started to add a bit of help documentation to make my own life easier.  I also found a few bugs that I fixed and did some improvements but I really tried not to diverge from the original functionality.

Now, since I find it a waste of resources if someone else has to go through the same efforts that I had to, I decided to make this package public.  The goal is also to submit the **TopDom** package to CRAN so that the TopDom method is properly archived for reproducible purposes.  In order to do this, the original authors agreed on releasing their original TopDom scripts under the GPL license, which is also the license of the **TopDom** package.

Having said this, please note that I don't intend to do user support for the TopDom method, how to use it, add new features, and so on.  This is simply because I don't have the resources to do that.  More importantly, I am not the best person to answer questions on the TopDom method and its implementation.  Instead, I recommend that you reach out to the original TopDom authors to get your questions answered.


## Archaeology

The original R script (versions 0.0.1 and 0.0.2) and the TopDom manual were made available on a website of the TopDom author (http<span/>://zhoulab.usc.edu/TopDom/).  As of 2019, that website is no longer available.  Instead, TopDom is now listed as a paragraph on <https://labs.dgsom.ucla.edu/zhou/pages/software> with a link to one of the author GitHub repository [jasminezhoulab/TopDom](https://github.com/jasminezhoulab/TopDom), which holds the TopDom manual and version 0.0.2 of the script but not version 0.0.1.



[R]: https://www.r-project.org/
[TopDom]: https://github.com/HenrikBengtsson/TopDom/
[TopDomData]: https://github.com/HenrikBengtsson/TopDomData/

## Installation
R package TopDom is available on [CRAN](https://cran.r-project.org/package=TopDom) and can be installed in R as:
```r
install.packages("TopDom")
```


### Pre-release version

To install the pre-release version that is available in Git branch `develop` on GitHub, use:
```r
remotes::install_github("HenrikBengtsson/TopDom", ref="develop")
```
This will install the package from source.  

<!-- pkgdown-drop-below -->

# @author : Hanjun Shin
# @brief : TopDom.R is a software package to identify topological domains for given Hi-C contact matrix. 

# @fn TopDom
# @param matrixFile : string, matrixFile Address,
# - Format = {chromosome, bin start, bin end, N numbers normalized value }
# - N * (N + 3), where N is the number of bins
# @param window.size :integer, number of bins to extend.
# @param out_binSignal : string, binSignal file address to write
# @param out_ext : string, ext file address to write
TopDom <- function( matrix.file, window.size, outFile=NULL, statFilter=T)
{
  print("#########################################################################")
  print("Step 0 : File Read ")
  print("#########################################################################")
  window.size = as.numeric(window.size)
  matdf = read.table(matrix.file, header=F)
  
  if( ncol(matdf) - nrow(matdf) == 3) {
    colnames(matdf) <- c("chr", "from.coord", "to.coord")
  } else if( ncol(matdf) - nrow(matdf) ==4 ) {
    colnames(matdf) <- c("id", "chr", "from.coord", "to.coord")
  } else {
    print("Unknwon Type of matrix file")
    return(0)
  }
  
  n_bins = nrow(matdf)
  mean.cf <- rep(0, n_bins)
  pvalue <- rep(1, n_bins)
  
  local.ext = rep(-0.5, n_bins)
  
    
  bins <- data.frame(id=1:n_bins, 
                     chr=matdf[, "chr"], 
                     from.coord=matdf[, "from.coord"], 
                     to.coord=matdf[, "to.coord"] )
  matrix.data <- as.matrix( matdf[, (ncol(matdf) - nrow(matdf)+1 ):ncol(matdf)] )
  
  print("-- Done!")
  print("Step 0 : Done !!")
  
    
  print("#########################################################################")
  print("Step 1 : Generating binSignals by computing bin-level contact frequencies")
  print("#########################################################################")
  ptm <- proc.time()
  for(i in 1:n_bins)
  {
    diamond = Get.Diamond.Matrix(mat.data=matrix.data, i=i, size=window.size)
    mean.cf[i] = mean(diamond)    
  }
  
  eltm = proc.time() - ptm
  print(paste("Step 1 Running Time : ", eltm[3]))
  print("Step 1 : Done !!")
  
  print("#########################################################################")
  print("Step 2 : Detect TD boundaries based on binSignals")
  print("#########################################################################")
  
  ptm = proc.time()
  gap.idx = Which.Gap.Region(matrix.data=matrix.data)
  proc.regions = Which.process.region(rmv.idx=gap.idx, n_bins=n_bins, min.size=3)
  
  for( i in 1:nrow(proc.regions))
  {
    start = proc.regions[i, "start"]
    end = proc.regions[i, "end"]
    
    print(paste("Process Regions from ", start, "to", end))
      
    local.ext[start:end] = Detect.Local.Extreme(x=mean.cf[start:end])
  }
  
  eltm = proc.time() - ptm
  print(paste("Step 2 Running Time : ", eltm[3]))
  print("Step 2 : Done !!")
  
  if(statFilter)
  {
    print("#########################################################################")
    print("Step 3 : Statistical Filtering of false positive TD boundaries")
    print("#########################################################################")
  
    ptm = proc.time()  
    print("-- Matrix Scaling....")
    scale.matrix.data = matrix.data
    for( i in 1:(2*window.size) )
    {
      #diag(scale.matrix.data[, i:n_bins] ) = scale( diag( matrix.data[, i:n_bins] ) )
      scale.matrix.data[ seq(1+(n_bins*i), n_bins*n_bins, 1+n_bins) ] = scale( matrix.data[ seq(1+(n_bins*i), n_bins*n_bins, 1+n_bins) ] )
    }
    
    print("-- Compute p-values by Wilcox Ranksum Test")
    for( i in 1:nrow(proc.regions))
    {
      start = proc.regions[i, "start"]
      end = proc.regions[i, "end"]
      
      print(paste("Process Regions from ", start, "to", end))
      
      pvalue[start:end] <- Get.Pvalue(matrix.data=scale.matrix.data[start:end, start:end], size=window.size, scale=1)
    }
    print("-- Done!")
    
    print("-- Filtering False Positives")
    local.ext[intersect( union(which( local.ext==-1), which(local.ext==-1)), which(pvalue<0.05))] = -2
    local.ext[which(local.ext==-1)] = 0
    local.ext[which(local.ext==-2)] = -1
    print("-- Done!")
    
    eltm = proc.time() - ptm
    print(paste("Step 3 Running Time : ", eltm[3]))
    print("Step 3 : Done!")
  } else pvalue = 0

  domains = Convert.Bin.To.Domain.TMP(bins=bins, 
                                  signal.idx=which(local.ext==-1), 
                                  gap.idx=which(local.ext==-0.5), 
                                  pvalues=pvalue, 
                                  pvalue.cut=0.05)
  bins = cbind(bins, 
               local.ext = local.ext,
               mean.cf = mean.cf, 
               pvalue = pvalue)
  
  if( !is.null(outFile) ) {
    print("#########################################################################")
    print("Writing Files")
    print("#########################################################################")
    
    outBinSignal =  paste(outFile, ".binSignal", sep="")
    print(paste("binSignal File :", outBinSignal) )
    write.table(bins, file=outBinSignal, quote=F, row.names=F, col.names=T, sep="\t")  
  
  
    outDomain = paste(outFile, ".domain", sep="")
    print(paste("Domain File :", outDomain) )
    write.table( domains, file=outDomain, quote=F, row.names=F, col.names=T, sep="\t")
  }
  
  print("Done!!")
  
  print("Job Complete !")
  return(list(binSignal=bins, domain=domains))
}

# @fn Get.Diamond.Matrix
# @param mat.data : N by N matrix, where each element indicate contact frequency
# @param i :integer, bin index
# @param size : integer, window size to expand from bin
# @retrun : matrix.
Get.Diamond.Matrix <- function(mat.data, i, size)
{
  n_bins = nrow( mat.data )
  if(i==n_bins) return(NA)
  
  lowerbound = max( 1, i-size+1 )
  upperbound = min( i+size, n_bins)
  
  return( mat.data[lowerbound:i, (i+1):upperbound] )
}

# @fn Which.process.region
# @param rmv.idx : vector of idx, remove index vector
# @param n_bins : total number of bins
# @param min.size : minimum size of bins
# @retrun : data.frame of proc.regions
Which.process.region <- function(rmv.idx, n_bins, min.size=3)
{
  gap.idx = rmv.idx
  
  proc.regions = data.frame(start=numeric(0), end=numeric(0))
  proc.set = setdiff(1:n_bins, gap.idx)
  n_proc.set = length(proc.set)
  
  i=1
  while(i < n_proc.set )
  {
    start = proc.set[i]
    j = i+1
    
    while(j <= n_proc.set)
    {
      if( proc.set[j] - proc.set[j-1] <= 1) j = j + 1
      else {
        proc.regions = rbind(proc.regions, c(start=start, end=proc.set[j-1]) )
        i = j
        break
      }
    }
    
    if(j >= n_proc.set ) {
      proc.regions = rbind(proc.regions, c(start=start, end=proc.set[j-1]) )
      break
    }
  }
  
  colnames(proc.regions) = c("start", "end")
  proc.regions <- proc.regions[ which( abs(proc.regions[,"end"] - proc.regions[, "start"]) >= min.size ), ]
  
  return(proc.regions)
}

# @fn Which.Gap.Region
# @param matrix.data : n by n matrix
# @return gap index
Which.Gap.Region <- function(matrix.data)
{
  n_bins = nrow(matrix.data)
  gap = rep(0, n_bins)
  
  i=1
  while(i < n_bins)
  {
    j = i + 1
    while( j <= n_bins)
    {
      if( sum( matrix.data[i:j, i:j]) == 0 ) {
        gap[i:j] = -0.5
        j = j+1
      } else break
    }
    
    i = j  
  }
  idx = which(gap == -0.5)
  return(idx)
}

# @fn Detect.Local.Extreme
# @param x : original signal to find local minima
# @return vector of local extrme, -1 if the index is local minimum, 1 if the index is local maxima, 0 otherwise.
Detect.Local.Extreme <- function(x)
{
  n_bins = length(x)
  ret = rep(0, n_bins)
  x[is.na(x)]=0
  
  if(n_bins <= 3)
  {
    ret[which.min(x)]=-1
    ret[which.max(x)]=1
    
    return(ret)
  }
  # Norm##################################################3
  new.point = Data.Norm(x=1:n_bins, y=x)
  x=new.point$y
  ##################################################
  cp = Change.Point(x=1:n_bins, y=x)
  
  if( length(cp$cp) <= 2 ) return(ret)
  for(i in 2:(length(cp$cp)-1))
  {
    if( x[cp$cp[i]] >= x[cp$cp[i]-1] && x[cp$cp[i]] >= x[cp$cp[i]+1] ) ret[cp$cp[i]] = 1
    else if(x[cp$cp[i]] < x[cp$cp[i]-1] && x[cp$cp[i]] < x[cp$cp[i]+1]) ret[cp$cp[i]] = -1
    
    min.val = min( x[ cp$cp[i-1] ], x[ cp$cp[i] ] )
    max.val = max( x[ cp$cp[i-1] ], x[ cp$cp[i] ] )
    
    if( min( x[cp$cp[i-1]:cp$cp[i]] ) < min.val ) ret[ cp$cp[i-1] - 1 + which.min( x[cp$cp[i-1]:cp$cp[i]] ) ] = -1
    if( max( x[cp$cp[i-1]:cp$cp[i]] ) > max.val ) ret[ cp$cp[i-1] - 1 + which.max( x[cp$cp[i-1]:cp$cp[i]] ) ] = 1
  }

  return(ret)
}

# @fn Data.Norm
# @param x : x axis vector
# @param x : y axis vector
# @return list of normalized x and y
Data.Norm <- function(x, y)
{
  ret.x = rep(0, length(x))
  ret.y = rep(0, length(y))
  
  ret.x[1] = x[1]
  ret.y[1] = y[1]
  
  diff.x = diff(x)
  diff.y = diff(y)
  
  scale.x = 1 / mean( abs(diff(x) ) )
  scale.y = 1 / mean( abs( diff(y) ) )
  
  #print(scale.x)
  #print(scale.y)
  
  for(i in 2:length(x))
  {
    ret.x[i] = ret.x[i-1] + (diff.x[i-1]*scale.x)
    ret.y[i] = ret.y[i-1] + (diff.y[i-1]*scale.y)
  }
  
  return(list(x=ret.x, y=ret.y))
}

# @fn Change.Point
# @param x : x axis vector
# @param x : y axis vector
# @return change point index in x vector,
# Note that the first and the last point will be always change point
Change.Point <- function( x, y )
{
  if( length(x) != length(y)) 
  {
    print("ERROR : The length of x and y should be the same")
    return(0)
  }
  
  n_bins <- length(x)
  Fv <- rep(NA, n_bins)
  Ev <- rep(NA, n_bins)
  cp <- 1
  
  i=1
  Fv[1]=0
  while( i < n_bins )
  {
    j=i+1
    Fv[j] = sqrt( (x[j]-x[i])^2 + (y[j] - y[i] )^2 )
    
    while(j<n_bins)
    {
      j=j+1
      k=(i+1):(j-1)
      Ev[j] = ( sum( abs( (y[j]-y[i] )*x[k] - (x[j] -x[i])*y[k] - (x[i]*y[j]) + (x[j]*y[i]) ) ) / sqrt( (x[j]-x[i])^2 + (y[j] - y[i] )^2 ) )
      Fv[j] = sqrt( (x[j]-x[i])^2 + (y[j] - y[i])^2 ) - ( sum( abs( (y[j]-y[i] )*x[k] - (x[j] -x[i])*y[k] - (x[i]*y[j]) + (x[j]*y[i]) ) ) / sqrt( (x[j]-x[i])^2 + (y[j] - y[i] )^2 ) )
      
      #################################################
      #Not Original Code
      if( is.na(Fv[j]) || is.na(Fv[j-1]) ) {
        j = j-1
        cp <- c(cp, j)
        break
      }
      ####################################################3
      if(Fv[j] < Fv[j-1] ) {
        j = j - 1
        cp <- c(cp, j )
        break
      }
    }
    i=j
  }
  
  cp <- c(cp, n_bins)
  
  return(list(cp=cp, objF=Fv, errF=Ev))
}

# @fn Get.Pvalue
# @param matrix.data : matrix
# @param size : size to extend
# @param scale : scale parameter if necessary. deprecated parameter
# @return computed p-value vector
Get.Pvalue <- function( matrix.data, size, scale=1 )
{
  n_bins = nrow(matrix.data)
  pvalue <- rep(1, n_bins)
  
  for( i in 1:(n_bins-1) )
  {
    dia = as.vector( Get.Diamond.Matrix2(matrix.data, i, size=size) )
    ups = as.vector( Get.Upstream.Triangle(matrix.data, i, size=size) )
    downs = as.vector( Get.Downstream.Triangle(matrix.data, i, size=size) )
    
    wil.test =  wilcox.test(x=dia*scale, y=c(ups, downs), alternative="less", exact=F)
    pvalue[i] = wil.test$p.value  
    
    #print(paste(i, "=", wil.test$p.value) )  
  }
  
  pvalue[ is.na(pvalue) ] = 1
  return(pvalue)
}

# @fn Get.Upstream.Triangle
# @param mat.data : matrix data
# @param i : bin index
# @param size : size of window to extend
# @return upstream triangle matrix
Get.Upstream.Triangle <- function(mat.data, i, size)
{
  n_bins = nrow(mat.data)
  
  lower = max(1, i-size)
  tmp.mat = mat.data[lower:i, lower:i]
  return( tmp.mat[ upper.tri( tmp.mat, diag=F ) ] )
}

# @fn Get.Downstream.Triangle
# @param mat.data : matrix data
# @param i : bin index
# @param size : size of window to extend
# @return downstream triangle matrix
Get.Downstream.Triangle <- function(mat.data, i, size)
{
  n_bins = nrow(mat.data)
  if(i==n_bins) return(NA)
  
  upperbound = min(i+size, n_bins)
  tmp.mat = mat.data[(i+1):upperbound, (i+1):upperbound]
  return( tmp.mat[ upper.tri( tmp.mat, diag=F ) ] )
}

# @fn Get.Diamond.Matrix2
# @param mat.data : matrix data
# @param i : bin index
# @param size : size of window to extend
# @return diamond matrix
Get.Diamond.Matrix2 <- function(mat.data, i, size)
{
  n_bins = nrow(mat.data)
  new.mat = matrix(rep(NA, size*size), nrow=size, ncol=size)
  
  for(k in 1:size)
  {
    if(i-(k-1) >= 1 && i < n_bins)
    {
      lower = min(i+1, n_bins)
      upper = min(i+size, n_bins)
      
      new.mat[size-(k-1), 1:(upper-lower+1)] = mat.data[i-(k-1), lower:upper]
    }
  }
  
  return(new.mat)
}
  
# @fn Convert.Bin.To.Domain
# @param bins : bin information
# @param signal.idx : signal index
# @param signal.idx : gap index
# @param pvalues : pvalue vector
# @param pvalue.cut : pvalue threshold
# @return dataframe storing domain information
Convert.Bin.To.Domain <- function(bins, signal.idx, gap.idx, pvalues=NULL, pvalue.cut=NULL)
{
  n_bins = nrow(bins)
  ret = data.frame(chr=character(0), from.id=numeric(0), from.coord=numeric(0), to.id=numeric(0), to.coord=numeric(0), tag=character(0), size=numeric(0))
  levels( x=ret[, "tag"] ) = c("domain", "gap", "boundary")
  
  rmv.idx = setdiff(1:n_bins, gap.idx)
  proc.region = Which.process.region(rmv.idx, n_bins, min.size=0)
  from.coord = bins[proc.region[, "start"], "from.coord"]
  n_procs = nrow(proc.region)
  gap = data.frame(chr=rep( bins[1, "chr"], n_procs), from.id=rep(0, n_procs), from.coord=from.coord, to.id=rep(0, n_procs), to.coord=rep(0, n_procs), tag=rep("gap", n_procs), size=rep(0, n_procs), stringsAsFactors=F)
  
  rmv.idx = union(signal.idx, gap.idx)
  proc.region = Which.process.region(rmv.idx, n_bins, min.size=0)
  n_procs = nrow(proc.region)
  from.coord = bins[proc.region[, "start"], "from.coord"]
  domain = data.frame(chr=rep( bins[1, "chr"], n_procs), from.id=rep(0, n_procs), from.coord=from.coord, to.id=rep(0, n_procs), to.coord=rep(0, n_procs), tag=rep("domain", n_procs), size=rep(0, n_procs), stringsAsFactors=F)
  
  rmv.idx = setdiff(1:n_bins, signal.idx)
  proc.region = as.data.frame( Which.process.region(rmv.idx, n_bins, min.size=1) )
  n_procs = nrow(proc.region)
  if(n_procs>0)
  {
    from.coord = bins[proc.region[, "start"]+1, "from.coord"]  
    boundary = data.frame(chr=rep( bins[1, "chr"], n_procs), from.id=rep(0, n_procs), from.coord=from.coord, to.id=rep(0, n_procs), to.coord=rep(0, n_procs), tag=rep("boundary", n_procs), size=rep(0, n_procs), stringsAsFactors=F)
    ret = rbind(ret, boundary)
  }
  
  ret = rbind(gap, domain)
  ret = ret[order(ret[,3]), ]
  
  ret[, "to.coord"] = c(ret[2:nrow(ret), "from.coord"], bins[n_bins, "to.coord"])
  ret[, "from.id"] = match( ret[, "from.coord"], bins[, "from.coord"] )
  ret[, "to.id"] = match(ret[, "to.coord"], bins[, "to.coord"])
  ret[, "size"] = ret[,"to.coord"]-ret[,"from.coord"]
  
  if(!is.null(pvalues) && !is.null(pvalue.cut))
  {
    for(i in 1:nrow(ret))
    {
      if(ret[i, "tag"]=="domain")
      {
        domain.bins.idx = ret[i, "from.id"]:ret[i, "to.id"]
        p.value.constr = which( pvalues[ domain.bins.idx ] < pvalue.cut )
        
        if( length(domain.bins.idx) == length(p.value.constr)) ret[i, "tag"] = "boundary"
      }
    }
  }
  
  new.bdr.set = data.frame(chr=character(0), from.id=numeric(0), from.coord=numeric(0), to.id=numeric(0), to.coord=numeric(0), tag=character(0), size=numeric(0))
  stack.bdr = data.frame(chr=character(0), from.id=numeric(0), from.coord=numeric(0), to.id=numeric(0), to.coord=numeric(0), tag=character(0), size=numeric(0))
  
  i=1
  while(i <= nrow(ret))
  {
    if( ret[i, "tag"] == "boundary" )
    {
      stack.bdr = rbind(stack.bdr, ret[i, ])
    } else if(nrow(stack.bdr)>0) {
      new.bdr = data.frame(chr=bins[1, "chr"], 
                           from.id = min( stack.bdr[, "from.id"]), 
                           from.coord=min(stack.bdr[, "from.coord"]),
                           to.id = max( stack.bdr[, "to.id"]),
                           to.coord=max(stack.bdr[, "to.coord"]),
                           tag="boundary",
                           size=max(stack.bdr[, "to.coord"]) - min(stack.bdr[, "from.coord"]))
      new.bdr.set = rbind(new.bdr.set, new.bdr)
      stack.bdr = data.frame(chr=character(0), from.id=numeric(0), from.coord=numeric(0), to.id=numeric(0), to.coord=numeric(0), tag=character(0), size=numeric(0))
    }
    
    i = i + 1
  }
  
  ret = rbind( ret[ ret[, "tag"]!="boundary", ], new.bdr.set )
  ret = ret[order(ret[, "to.coord"]), ]
  
  return(ret)
}


# @fn Convert.Bin.To.Domain
# @param bins : bin information
# @param signal.idx : signal index
# @param signal.idx : gap index
# @param pvalues : pvalue vector
# @param pvalue.cut : pvalue threshold
# @return dataframe storing domain information
Convert.Bin.To.Domain.TMP <- function(bins, signal.idx, gap.idx, pvalues=NULL, pvalue.cut=NULL)
{
  n_bins = nrow(bins)
  ret = data.frame(chr=character(0), from.id=numeric(0), from.coord=numeric(0), to.id=numeric(0), to.coord=numeric(0), tag=character(0), size=numeric(0))
  levels( x=ret[, "tag"] ) = c("domain", "gap", "boundary")
  
  rmv.idx = setdiff(1:n_bins, gap.idx)
  proc.region = Which.process.region(rmv.idx, n_bins, min.size=0)
  from.coord = bins[proc.region[, "start"], "from.coord"]
  n_procs = nrow(proc.region)
  gap = data.frame(chr=rep( bins[1, "chr"], n_procs), from.id=rep(0, n_procs), from.coord=from.coord, to.id=rep(0, n_procs), to.coord=rep(0, n_procs), tag=rep("gap", n_procs), size=rep(0, n_procs), stringsAsFactors=F)
  
  rmv.idx = union(signal.idx, gap.idx)
  proc.region = Which.process.region(rmv.idx, n_bins, min.size=0)
  n_procs = nrow(proc.region)
  from.coord = bins[proc.region[, "start"], "from.coord"]
  domain = data.frame(chr=rep( bins[1, "chr"], n_procs), from.id=rep(0, n_procs), from.coord=from.coord, to.id=rep(0, n_procs), to.coord=rep(0, n_procs), tag=rep("domain", n_procs), size=rep(0, n_procs), stringsAsFactors=F)
  
  rmv.idx = setdiff(1:n_bins, signal.idx)
  proc.region = as.data.frame( Which.process.region(rmv.idx, n_bins, min.size=1) )
  n_procs = nrow(proc.region)
  if(n_procs>0)
  {
    from.coord = bins[proc.region[, "start"]+1, "from.coord"]  
    boundary = data.frame(chr=rep( bins[1, "chr"], n_procs), from.id=rep(0, n_procs), from.coord=from.coord, to.id=rep(0, n_procs), to.coord=rep(0, n_procs), tag=rep("boundary", n_procs), size=rep(0, n_procs), stringsAsFactors=F)
    ret = rbind(ret, boundary)
  }
  
  ret = rbind(gap, domain)
  ret = ret[order(ret[,3]), ]
  
  ret[, "to.coord"] = c(ret[2:nrow(ret), "from.coord"], bins[n_bins, "to.coord"])
  ret[, "from.id"] = match( ret[, "from.coord"], bins[, "from.coord"] )
  ret[, "to.id"] = match(ret[, "to.coord"], bins[, "to.coord"])
  ret[, "size"] = ret[,"to.coord"]-ret[,"from.coord"]
  
  if(!is.null(pvalues) && !is.null(pvalue.cut))
  {
    for(i in 1:nrow(ret))
    {
      if(ret[i, "tag"]=="domain")
      {
        domain.bins.idx = ret[i, "from.id"]:ret[i, "to.id"]
        p.value.constr = which( pvalues[ domain.bins.idx ] < pvalue.cut )
        
        if( length(domain.bins.idx) == length(p.value.constr)) ret[i, "tag"] = "boundary"
      }
    }
  }
  
  return(ret)
}

#' Easy Access to the Original TopDom 0.0.1 and 0.0.2 Implementations
#'
#' @param version A version string.
#'
#' @return An environment containing the legacy TopDom API.
#'
#' @examples
#' TopDom::legacy("0.0.2")$TopDom
#' TopDom::legacy("0.0.1")$Detect.Local.Extreme
#' 
#' @export
legacy <- function(version = c("0.0.1", "0.0.2")) {
  if (version == "0.0.1") {
    api <- environment(TopDom_0.0.1)
  } else if (version == "0.0.2") {
    api <- environment(TopDom_0.0.2)
  } else {
    stop("Unknown TopDom legacy version: ", sQuote(version))
  }
  api
}

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")
pathname <- file.path(path, sprintf("nij.%s.gz", chr))

data <- readHiC(pathname, chr = chr, binSize = 40e3)
data <- data[1:200]

tds_5 <- TopDom(data, window.size = 5L)
tds_6 <- TopDom(data, window.size = 6L)

## overlapScores() with debug output ------------------------------------
msg <- utils::capture.output(
  ov <- overlapScores(tds_6, reference = tds_5, debug = TRUE),
  type = "message"
)
stopifnot(inherits(ov, "TopDomOverlapScores"), any(grepl("Chromosome", msg)))

## Self-overlap should be perfect (score 1) for every reference domain --
ov_self <- overlapScores(tds_5, reference = tds_5)
scores <- ov_self[[chr]]$best_score
stopifnot(all(scores[!is.na(scores)] == 1))

## print.TopDomOverlapScores() ----------------------------------------
out <- utils::capture.output(print(ov_self))
stopifnot(any(grepl("Chromosomes:", out)),
          any(grepl("best scores", out)))

## as_tibble.TopDomOverlapScores() -----------------------------------
tbl <- tibble::as_tibble(ov_self)
stopifnot(
  inherits(tbl, "data.frame"),
  all(c("chromosome", "best_score", "best_length", "best_set") %in% names(tbl)),
  nrow(tbl) == nrow(tds_5$domain)
)

## Non-TopDom inputs are rejected ----------------------------------
res <- tryCatch(overlapScores(tds_5, reference = list()), error = identity)
stopifnot(inherits(res, "error"))
res <- tryCatch(overlapScores("x", reference = tds_5), error = identity)
stopifnot(inherits(res, "error"))

#' @importFrom utils str
#' @export
print.TopDomData <- function(x, ...) {
  cat(sprintf("%s:\n", class(x)))
  cat("bins:\n")
  str(x$bins)
  cat("counts:\n")
  str(x$counts)
}

#' @export
dim.TopDomData <- function(x) {
  dim(x$counts)
}

#' @export
`[.TopDomData` <- function(x, i, ...) {
  structure(list(
    bins = x$bins[i, , drop = FALSE],
    counts = x$counts[i, i, drop = FALSE]
  ), class = "TopDomData")
}

#' @importFrom graphics image
#' @export
image.TopDomData <- function(x, transform = log2, ...) {
  image(transform(x$counts), ...)
}

library("TopDom")

if (requireNamespace("ggplot2", quietly = TRUE)) {
  path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
  chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")

  pathname <- file.path(path, sprintf("nij.%s.gz", chr))
  data <- readHiC(pathname, chr = chr, binSize = 40e3)
  data <- data[1:150]

  fit <- TopDom(data, window.size = 5L)
  domains <- subset(fit$domain, tag == "domain")
  td <- domains[which.max(domains$size), ]

  ## ggCountHeatmap() - default arguments
  gg <- ggCountHeatmap(data)
  stopifnot(inherits(gg, "ggplot"))
  b <- ggplot2::ggplot_build(gg)
  stopifnot(is.list(b$data), nrow(b$data[[1]]) > 0L)

  ## ggCountHeatmap() - custom transform and partial 'colors' (fills in defaults)
  gg2 <- ggCountHeatmap(data, transform = function(x) x, colors = c(high = "red"))
  stopifnot(inherits(gg2, "ggplot"))
  invisible(ggplot2::ggplot_build(gg2))

  ## ggDomain() - default 'dx' derived from 'delta'/'vline'
  layer <- ggDomain(td)
  stopifnot(inherits(layer, "Layer") || inherits(layer, "LayerInstance"))
  params <- attr(layer, "gg_params")
  stopifnot(
    is.list(params),
    params$width == td$to.coord - td$from.coord,
    params$dx == params$delta * params$width + params$vline
  )
  gg3 <- gg + layer
  stopifnot(inherits(gg3, "ggplot"))

  ## ggDomain() - explicit 'dx'
  layer2 <- ggDomain(td, dx = 1000, color = "black", size = 1)
  stopifnot(attr(layer2, "gg_params")$dx == 1000)

  ## ggDomainLabel() - returns something addable to a ggplot
  lab <- ggDomainLabel(td)
  gg4 <- gg + lab
  stopifnot(inherits(gg4, "ggplot"))
  invisible(ggplot2::ggplot_build(gg4))

  lab2 <- ggDomainLabel(td, fmt = "%s [%.1f-%.1f]", rot = 0, dx = 10, cex = 1)
  stopifnot(inherits(gg + lab2, "ggplot"))
}

library("TopDom")

options(TopDom.debug = FALSE)

## A tiny synthetic contact matrix -------------------------------------------
n <- 6L
set.seed(42L)
counts <- matrix(as.double(sample(0:9, size = n * n, replace = TRUE)), nrow = n)
counts[lower.tri(counts)] <- t(counts)[lower.tri(counts)]  ## symmetric
from <- seq(from = 0L, by = 40000L, length.out = n)
to <- from + 40000L

tmpdir <- tempfile("readHiC-")
dir.create(tmpdir)

write_ws <- function(df, file) {
  utils::write.table(df, file = file, sep = "\t", quote = FALSE,
                     row.names = FALSE, col.names = FALSE)
  file
}

## (1) N-by-(N+3): chr, from.coord, to.coord, <counts> ----------------------
f3 <- write_ws(data.frame(chr = "chr1", from = from, to = to, counts),
               file.path(tmpdir, "mat3.txt"))
d3 <- readHiC(f3)
str(d3)
stopifnot(
  inherits(d3, "TopDomData"),
  identical(dim(d3), c(n, n)),
  identical(d3$bins$id, seq_len(n)),
  all(d3$bins$chr == "chr1"),
  is.numeric(d3$counts),
  all(d3$counts == counts)
)

## (2) N-by-(N+4): id, chr, from.coord, to.coord, <counts> ----------------
f4 <- write_ws(data.frame(id = seq_len(n), chr = "chr1", from = from, to = to, counts),
               file.path(tmpdir, "mat4.txt"))
d4 <- readHiC(f4)
stopifnot(inherits(d4, "TopDomData"), identical(dim(d4), c(n, n)))
stopifnot(all(d4$counts == counts))

## (3) Unknown format -> error -------------------------------------------
f5 <- write_ws(data.frame(a = 1:n, b = 1:n, chr = "chr1", from = from, to = to, counts),
               file.path(tmpdir, "mat5.txt"))
res <- tryCatch(readHiC(f5), error = identity)
stopifnot(inherits(res, "error"), grepl("Unknown format", conditionMessage(res)))

## (4) Non-existing file -> error --------------------------------------
res <- tryCatch(readHiC(file.path(tmpdir, "does-not-exist.txt")), error = identity)
stopifnot(inherits(res, "error"))

## (5) Pure N-by-N count matrix + (chr, binSize) ------------------------
fm <- write_ws(as.data.frame(counts), file.path(tmpdir, "matN.txt"))
dm <- readHiC(fm, chr = "chrZ", binSize = 40e3)
stopifnot(
  inherits(dm, "TopDomData"),
  identical(dim(dm), c(n, n)),
  all(dm$bins$chr == "chrZ"),
  all(dm$bins$from.coord == from),
  all(dm$bins$to.coord == to)
)

## ... with debug output enabled
dm_dbg <- readHiC(fm, chr = "chrZ", binSize = 40e3,
                  debug = TRUE)
stopifnot(identical(dim(dm_dbg), c(n, n)))

## (6) 'bins' subsetting of a pure count matrix -----------------------
keep <- c(1L, 3L, 5L)
db <- readHiC(fm, chr = "chrZ", binSize = 40e3, bins = keep)
stopifnot(
  identical(dim(db), c(length(keep), length(keep))),
  identical(db$bins$id, keep),
  all(db$counts == counts[keep, keep])
)

## Out-of-range and non-positive 'bins' -> error
res <- tryCatch(readHiC(fm, chr = "chrZ", binSize = 40e3, bins = c(1L, 999L)),
                error = identity)
stopifnot(inherits(res, "error"), grepl("out of range", conditionMessage(res)))

res <- tryCatch(readHiC(fm, chr = "chrZ", binSize = 40e3, bins = c(0L, 2L)),
                error = identity)
stopifnot(inherits(res, "error"), grepl("non-positive", conditionMessage(res)))

## Invalid 'debug' argument -> error
res <- tryCatch(readHiC(f3, debug = "yes"), error = identity)
stopifnot(inherits(res, "error"))

unlink(tmpdir, recursive = TRUE)

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")
pathname <- file.path(path, sprintf("nij.%s.gz", chr))

data <- readHiC(pathname, chr = chr, binSize = 40e3)
data <- data[1:200]

## Reference fit -----------------------------------------------------------
fit <- TopDom(data, window.size = 5L)
stopifnot(inherits(fit, "TopDom"))
params <- attr(fit, "parameters")
stopifnot(params$window.size == 5L, isTRUE(params$statFilter))

## print.TopDom() --------------------------------------------------------
out <- utils::capture.output(print(fit))
stopifnot(length(out) > 0L)

## dim.TopDom() and [.TopDom -------------------------------------------
stopifnot(identical(dim(fit), dim(fit$domain)))
fit_sub <- fit[1:3, ]
stopifnot(inherits(fit_sub, "TopDom"), nrow(fit_sub$domain) == 3L)

## print.TopDom() when no parameters are attached
fit_noparam <- fit
attr(fit_noparam, "parameters") <- NULL
out <- utils::capture.output(print(fit_noparam))
stopifnot(any(grepl("N/A", out)))

## A larger window size yields 'boundary' domains, exercising the
## boundary-collapsing code path in Convert.Bin.To.Domain.TMP()
fit_w7 <- TopDom(data, window.size = 7L)
stopifnot("boundary" %in% fit_w7$domain$tag)
fit_w7_nf <- TopDom(data, window.size = 7L, statFilter = FALSE)
stopifnot(inherits(fit_w7_nf, "TopDom"))

## statFilter = FALSE branch -------------------------------------------
fit_nf <- TopDom(data, window.size = 5L, statFilter = FALSE)
stopifnot(
  inherits(fit_nf, "TopDom"),
  isFALSE(attr(fit_nf, "parameters")$statFilter),
  all(fit_nf$binSignal$pvalue == 0)
)

## Passing a file path as 'data' (dispatches to readHiC() internally) ---
fit_chr <- TopDom(pathname, window.size = 5L, chr = chr, binSize = 40e3,
                  bins = 1:200)
stopifnot(inherits(fit_chr, "TopDom"))
stopifnot(identical(fit_chr$domain, fit$domain))

## debug = TRUE exercises the messaging code paths --------------------
msg <- utils::capture.output(
  fit_dbg <- TopDom(data, window.size = 5L, debug = TRUE),
  type = "message"
)
stopifnot(inherits(fit_dbg, "TopDom"), any(grepl("Step 1", msg)))
stopifnot(identical(fit_dbg$domain, fit$domain))

## outFile = <prefix> writes .binSignal, .domain and .bed files -------
tmp <- tempfile("TopDom-")
fit_out <- TopDom(data, window.size = 5L, outFile = tmp)
for (ext in c(".binSignal", ".domain", ".bed")) {
  f <- paste0(tmp, ext)
  stopifnot(utils::file_test("-f", f), file.size(f) > 0L)
}
bed <- utils::read.table(paste0(tmp, ".bed"), sep = "\t", header = FALSE)
stopifnot(nrow(bed) == nrow(fit_out$bed))
unlink(paste0(tmp, c(".binSignal", ".domain", ".bed")))

## Corner case: input with no processable region -----------------------
## When every non-gap run is shorter than the minimum region size, the
## internal 'idxs' loop variable is never created.  TopDom() used to warn
## "object 'idxs' not found" from rm(list = "idxs"); make sure it does not.
n <- 10L
z <- matrix(0, nrow = n, ncol = n)
diag(z)[c(1L, n)] <- 5  ## isolated signal, no run of >= 3 bins
edge <- structure(list(
  bins = data.frame(id = seq_len(n), chr = "chr1",
                    from.coord = (0:(n - 1)) * 40000L,
                    to.coord = (1:n) * 40000L,
                    stringsAsFactors = FALSE),
  counts = z
), class = "TopDomData")
warns <- NULL
fit_edge <- withCallingHandlers(
  TopDom(edge, window.size = 3L, statFilter = FALSE),
  warning = function(w) {
    warns <<- c(warns, conditionMessage(w))
    invokeRestart("muffleWarning")
  }
)
stopifnot(inherits(fit_edge, "TopDom"))
stopifnot(!any(grepl("object 'idxs' not found", warns)))

## Corner case: extremely sparse data and statFilter = TRUE -----------
## Standardizing an all-zero near-diagonal band yields non-finite values
## that used to surface as an obscure wilcox.test() error
## <https://github.com/HenrikBengtsson/TopDom/issues/17>.  TopDom() should
## now fail early with an informative message that mentions statFilter.
m <- 40L
sparse <- structure(list(
  bins = data.frame(id = seq_len(m), chr = "22",
                    from.coord = (seq_len(m) - 1L) * 500000L,
                    to.coord = seq_len(m) * 500000L,
                    stringsAsFactors = FALSE),
  counts = { z <- matrix(0, nrow = m, ncol = m); diag(z) <- 5; z }
), class = "TopDomData")
res <- tryCatch(TopDom(sparse, window.size = 5L), error = identity)
stopifnot(
  inherits(res, "error"),
  grepl("statFilter", conditionMessage(res)),
  grepl("sparse", conditionMessage(res))
)

## Corner case: matrix too large for statFilter index arithmetic -------
## With more than sqrt(.Machine$integer.max) bins, the internal
## 'n_bins * n_bins' overflows R's integer range and seq() used to fail
## with an obscure "'to' must be a finite number" error
## <https://github.com/HenrikBengtsson/TopDom/issues/10>.  TopDom() should
## now fail early (before touching 'counts') with an informative message,
## but only when statFilter = TRUE.
N <- as.integer(sqrt(.Machine$integer.max)) + 1L
big <- structure(list(
  bins = data.frame(id = seq_len(N), chr = "1",
                    from.coord = (seq_len(N) - 1) * 1000,
                    to.coord = seq_len(N) * 1000),
  counts = matrix(0, nrow = 2L, ncol = 2L)  ## deliberately tiny
), class = "TopDomData")
res <- tryCatch(TopDom(big, window.size = 5L), error = identity)
stopifnot(
  inherits(res, "error"),
  grepl("statFilter", conditionMessage(res)),
  grepl("overflow", conditionMessage(res)),
  grepl(as.character(N), conditionMessage(res), fixed = TRUE)
)
## statFilter = FALSE must not raise this particular error
res <- tryCatch(TopDom(big, window.size = 5L, statFilter = FALSE), error = identity)
stopifnot(!grepl("overflow R's integer range", conditionMessage(res)))

## Invalid arguments -----------------------------------------------------
res <- tryCatch(TopDom(data, window.size = -1L), error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch(TopDom(data, window.size = 5L, debug = NA), error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch(TopDom(42, window.size = 5L), error = identity)
stopifnot(inherits(res, "error"))

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")

## From Supplementary Materials of TopDom article
## Source: http://zhoulab.usc.edu/TopDom/topdom_sup.htm
message("Loading truth ...")
pathname <- file.path(path, sprintf("mESC_5w_%s.nij.HindIII.comb.40kb.domain", chr))
truth <- read.table(pathname, sep = "\t", header = TRUE,
                    colClasses = c("factor", "integer", "numeric", "integer",
                                   "numeric", "factor", "numeric"))
str(truth)

## Original count data
message("Loading count data ...")
pathname <- file.path(path, sprintf("nij.%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3)
str(data)

## print()
print(data)

## dim()
print(dim(data))

## Subsetting via [()
data_s <- data[101:200]
str(data_s)
stopifnot(identical(dim(data_s), c(100L, 100L)))

## image()
image(data_s)

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")

## From Supplementary Materials of TopDom article
## Source: http://zhoulab.usc.edu/TopDom/topdom_sup.htm
message("Loading truth ...")
pathname <- file.path(path, sprintf("mESC_5w_%s.nij.HindIII.comb.40kb.domain", chr))
truth <- read.table(pathname, sep = "\t", header = TRUE,
                    colClasses = c("factor", "integer", "numeric", "integer",
                                   "numeric", "factor", "numeric"))
str(truth)

## Original count data
message("Loading count data ...")
pathname <- file.path(path, sprintf("nij.%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3)
str(data)

## Find topological domains using TopDom method
message("TopDom() ...")
fit <- TopDom(data, window.size = 5L)
str(fit$domain)

message("TopDom() v0.0.1 ...")
fit1 <- TopDom::legacy("0.0.1")$TopDom(data, window.size = 5L)
str(fit1$domain)

message("TopDom() v0.0.2 ...")
fit2 <- TopDom::legacy("0.0.2")$TopDom(data, window.size = 5L)
str(fit2$domain)

if (requireNamespace("diffobj", quietly = TRUE)) {
  message("TopDom v0.0.1 versus published results ...")
  diff <- diffobj::diffPrint(fit1$domain, truth,
                             extra = list(row.names = FALSE))
  print(diff)

  message("TopDom v0.0.2 versus published results ...")
  diff <- diffobj::diffPrint(fit2$domain, truth,
                             extra = list(row.names = FALSE))
  print(diff)

  message("TopDom v0.0.2 versus TopDom v0.0.1 results ...")
  diff <- diffobj::diffPrint(fit2$domain, fit1$domain,
                             extra = list(row.names = FALSE))
  print(diff)

  message("TopDom (package version) versus TopDom v0.0.2 results ...")
  diff <- diffobj::diffPrint(fit$domain, fit2$domain,
                             extra = list(row.names = FALSE))
  print(diff)
}

#' @importFrom utils capture.output
mcat <- function(...) {
  msg <- paste0(...)
  cat(msg, "\n", sep = "", file = stderr())
}


#' Asserts the Truth of R Expressions
#'
#' @param \dots Zero or more \R expressions to be asserted to be TRUE.
#'
#' @return Nothing.
#'
#' @details
#' A bare bone, faster version of [base::stopifnot].
#'
#' @keywords internal
stop_if_not <- function(...) {
  res <- list(...)
  for (ii in seq_along(res)) {
    res_ii <- .subset2(res, ii)
    if (length(res_ii) != 1L || is.na(res_ii) || !res_ii) {
        mc <- match.call()
        call <- deparse(mc[[ii + 1]], width.cutoff = 60L)
        if (length(call) > 1L) call <- paste(call[1L], "....")
        stop(sprintf("%s is not TRUE", sQuote(call)),
             call. = FALSE, domain = NA)
    }
  }
  invisible()
}

#' Calculates Overlap Scores Between Two Sets of Topological Domains
#' 
#' @param a,reference Topological domain (TD) set \eqn{A} and TD reference
#' set \eqn{R} both in a format as returned by [TopDom()].
#'
#' @param debug If `TRUE`, debug output is produced.
#' 
#' @return
#' Returns a named list of class `TopDomOverlapScores`, where the names
#' correspond to the chromosomes in domain reference set \eqn{R}.
#' Each of these chromosome elements contains a data.frame with fields:
#'
#' * `chromosome`  - \eqn{D_{R,c}} character strings
#' * `best_score`  - \eqn{D_{R,c}} numerics in \eqn{[0,1]}
#' * `best_length` - \eqn{D_{R,c}} positive integers
#' * `best_set`    - list of \eqn{D_{R,c}} index vectors
#'
#' where \eqn{D_{R,c}} is the number of TDs in reference set \eqn{R} on
#' chromosome \eqn{c}.  If a TD in reference \eqn{R} is not a `"domain"`,
#' then the corresponding `best_score` and `best_length` values are
#' `NA_real_` and `NA_integer_`, respectively, while `best_set` is an empty
#' list.
#'
#' @details
#' The _overlap score_, \eqn{overlap(A', r_i)}, represents how well a
#' _consecutive_ subset \eqn{A'} of topological domains (TDs) in \eqn{A}
#' overlap with topological domain \eqn{r_i} in reference set \eqn{R}.
#' For each reference TD \eqn{r_i}, the _best match_ \eqn{A'_{max}} is
#' identified, that is, the \eqn{A'} subset that maximize
#' \eqn{overlap(A', r_i)}.
#' For exact definitions, see Page 8 in Shin et al. (2016).
#'
#' Note that the overlap score is an asymmetric score, which means that
#' `overlapScores(a, b) != overlapScores(b, a)`.
#'
#' @section Warning - This might differ not be the correct implementation:
#' The original TopDom scripts do not provide an implementation for
#' calculating overlap scores.  Instead, the implementation of
#' `TopDom::overlapScores()` is based on the textual description of
#' overlap scores provided in Shin et al. (2016).  It is not known if this
#' is the exact same algorithm and implementation as the authors of the
#' TopDom article used.
#'
#' @example incl/overlapScores.R
#' 
#' @references
#' * Shin et al.,
#'   TopDom: an efficient and deterministic method for identifying
#'   topological domains in genomes,
#'   _Nucleic Acids Research_, 44(7): e70, April 2016.
#'   doi: \doi{10.1093/nar/gkv1505},
#'   PMCID: PMC4838359,
#'   PMID: 26704975
#'
#' @author Henrik Bengtsson - based on the description in Shin et al. (2016).
#' 
#' @seealso [TopDom].
#'
#' @importFrom tibble as_tibble tibble
#' @export
overlapScores <- function(a, reference, debug = getOption("TopDom.debug", FALSE)) {
  stopifnot(inherits(reference, "TopDom"), inherits(a, "TopDom"))
  stopifnot(is.logical(debug), length(debug) == 1L, !is.na(debug))

  ## Extract the 'domain' fields from the 'reference' and the 'a' TopDom objects
  domains_R <- reference$domain
  domains_A <- a$domain
  
  chromosomes <- unique(domains_R$chr)
  scores <- vector("list", length = length(chromosomes))
  names(scores) <- chromosomes
  for (chr in chromosomes) {
    if (debug) message(sprintf("Chromosome %s ...", chr))
    doms_R <- domains_R[domains_R$chr == chr, ]
    doms_A <- domains_A[domains_A$chr == chr, ]
    scores_chr <- overlapScoresOneChromosome(doms_A, doms_R = doms_R, debug = debug)
    scores_chr <- as_tibble(cbind(tibble(chromosome = chr), scores_chr))
    scores[[chr]] <- scores_chr
    if (debug) message(sprintf("Chromosome %s ... done", chr))
  }

  class(scores) <- c("TopDomOverlapScores", class(scores))
  scores
}


#' @importFrom tibble as_tibble
overlapScoresOneChromosome <- function(doms_A, doms_R, debug = getOption("TopDom.debug", FALSE)) {
  stopifnot(is.logical(debug), length(debug) == 1L, !is.na(debug))
  ## FIXME: Assert that A and reference R are sorted by (chr, pos)

  dtags <- diff(c(0L, as.integer(doms_A$tag == "domain"), 0L))
  sets <- data.frame(
    first = which(dtags == +1L),
    last  = which(dtags == -1L) - 1L,
    stringsAsFactors = FALSE
  )
  sets$from.coord <- doms_A$from.coord[sets$first]
  sets$to.coord <- doms_A$to.coord[sets$last]

  doms_R$length <- as.integer(doms_R$to.coord - doms_R$from.coord)
  doms_A$length <- as.integer(doms_A$to.coord - doms_A$from.coord)

  best_scores <- rep(NA_real_, times = nrow(doms_R))
  best_lengths <- rep(NA_integer_, times = nrow(doms_R))
  best_sets <- vector("list", length = nrow(doms_R))
  idxs_td <- which(doms_R$tag == "domain")
  for (ii in seq_along(idxs_td)) {
    idx_td <- idxs_td[ii]
    if (debug) message(sprintf("TD \"domain\" #%d of %d ...", ii, length(idxs_td)))
    td_R <- doms_R[idx_td, ]
    
    ## Identify sets to consider
    ## Q. How many sets can match this? [0,1], [0,2], [0,3], or even more?
    sets_t <- sets[(sets$to.coord >= td_R$from.coord &
                    sets$from.coord <= td_R$to.coord), ]

    best_score <- 0.0
    best_set <- integer(0L)
    
    ## For each possible A' set ...
    for (kk in seq_len(nrow(sets_t))) {
      set <- sets_t[kk,]
      doms <- doms_A[set$first:set$last,]
      doms$cap <- doms$length
      
      ## TD in A' that is overlapping part of the beginning of reference R 
      before <- which(doms$from.coord <= td_R$from.coord)
      if (length(before) > 0L) {
        before <- before[length(before)]
        doms$cap[before] <- doms$to.coord[before] - td_R$from.coord
      }
      
      ## TD in A' that is overlapping part of the end of reference R
      after <- which(doms$to.coord >= td_R$to.coord)
      if (length(after) > 0L) {
        after <- after[1L]
        doms$cap[after] <- td_R$to.coord - doms$from.coord[after]
      }
      
      ## TDs in A' that are strictly overlapping with reference R
      is_inside <- (doms$from.coord >= td_R$from.coord &
                    doms$to.coord <= td_R$to.coord)
      idxs_t <- c(before, which(is_inside), after)
      n <- length(idxs_t)
      stop_if_not(n > 0L)
      caps <- doms$cap[idxs_t]
      stop_if_not(length(caps) > 0L, any(is.finite(caps)))
      stop_if_not(!anyNA(caps))
      cups <- doms$length[idxs_t]
      stop_if_not(length(cups) > 0L, any(is.finite(cups)))
      if (n == 1L) {
        idxs_u <- list(1L)
        max_score <- caps / cups
      } else if (n == 2L) {
        idxs_u <- list(1L, 1:2, 2L)
      } else if (n >= 3L) {
        idxs_u <- list(1L, 1L:(n-1L), 2L:(n-1L), 2L:n, n)
      }
      stop_if_not(!anyNA(idxs_u))
      scores <- sapply(idxs_u, FUN = function(idxs) {
        sum(caps[idxs]) / sum(cups[idxs])
      })
      stop_if_not(any(is.finite(scores)))
      max_idx <- which.max(scores)

      max_score <- scores[max_idx]
      if (max_score > best_score) {
        best_score <- max_score
        best_set <- idxs_u[[max_idx]]
      }
    } ## for (kk ...)

    ## Sanity checks
    stop_if_not(length(best_score) == 1L, length(td_R$length) == 1L)

    best_scores[ii] <- best_score
    best_lengths[ii] <- td_R$length
    best_sets[[ii]] <- best_set
    
    if (debug) message(sprintf("TD \"domain\" #%d of %d ... done", ii, length(idxs_td)))
  } ## for (ii ...)

  ## Sanity checks
  stop_if_not(length(best_scores) == nrow(doms_R),
              length(best_lengths) == nrow(doms_R),
              length(best_sets) == nrow(doms_R))

  res <- data.frame(best_score = best_scores, best_length = best_lengths, stringsAsFactors = FALSE)
  res$best_set <- best_sets

  res
} ## overlapScoresOneChromosome()


#' @importFrom tibble as_tibble
#' @export
as_tibble.TopDomOverlapScores <- function(x, ...) {
  do.call(rbind, args = x)
}


#' @export
print.TopDomOverlapScores <- function(x, ...) {
  cat(sprintf("%s:\n", paste(class(x), collapse = ", ")))

  cat(sprintf("Chromosomes: [n = %d] %s\n",
              length(x), paste(sQuote(names(x)), collapse = ", ")))

  lengths <- lapply(x, FUN = `[[`, "best_length")
  lengths[["whole genome"]] <- unlist(lengths, use.names = FALSE)
  cat("Summary of reference domain lengths:\n")
  t <- t(sapply(lengths, FUN = function(x) {
    c(summary(x), count = length(x))
  }))
  print(t)
  
  scores <- lapply(x, FUN = `[[`, "best_score")
  scores[["whole genome"]] <- unlist(scores, use.names = FALSE)
  cat("Summary of best scores:\n")
  t <- t(sapply(scores, FUN = function(x) {
    c(summary(x), count = length(x))
  }))
  
  print(t)
}

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")

## From Supplementary Materials of TopDom article
## Source: http://zhoulab.usc.edu/TopDom/topdom_sup.htm
message("Loading truth ...")
pathname <- file.path(path, sprintf("mESC_5w_%s.nij.HindIII.comb.40kb.domain", chr))
truth <- read.table(pathname, sep = "\t", header = TRUE,
                    colClasses = c("factor", "integer", "numeric", "integer",
                                   "numeric", "factor", "numeric"))
str(truth)

## Original count data
message("Loading count data ...")
pathname <- file.path(path, sprintf("nij.%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3)
str(data)

## Find topological domains using TopDom method
message("TopDom() ...")
fit <- TopDom(data, window.size = 5L)
print(fit)
str(fit$domain)

if (requireNamespace("diffobj", quietly = TRUE)) {
  message("TopDom versus published results ...")
  diff <- diffobj::diffPrint(fit$domain, truth,
                             extra = list(row.names = FALSE))
  print(diff)
}

## The largest domain found
td <- subset(subset(fit$domain, tag == "domain"), size == max(size))
stopifnot(nrow(td) == 1L)

data_s <- subsetByRegion(data, region = td, margin = 1/2)
print(data_s)
image(data_s)

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")

## Original count data
pathname <- file.path(path, sprintf("nij.%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3)
str(data)

## Find topological domains using TopDom method for two window sizes
tds_5 <- TopDom(data, window.size = 5L)
tds_6 <- TopDom(data, window.size = 6L)

## Overlap scores (in both directions)
overlap_56 <- overlapScores(tds_6, reference = tds_5)
print(overlap_56)
overlap_65 <- overlapScores(tds_5, reference = tds_6)
print(overlap_65)

if (FALSE) {  
  ## Find topological domains as a function of windows size
  window.size <- c(5L, 6L, 8L, 10L, 12L, 14L, 16L, 18L, 20L)
  tds <- lapply(window.size, FUN = function(w) TopDom(data, window.size = w))
  names(tds) <- sprintf("window.size=%d", window.size)

  ## Overlap scores relative to the first window.size
  overlaps <- lapply(tds, FUN = overlapScores, reference = tds[[1]])
  print(overlaps)

  scores <- lapply(overlaps, FUN = function(overlaps) {
    unlist(lapply(overlaps, FUN = `[[`, "best_score"), use.names = FALSE)
  })

  avg_scores <- sapply(scores, FUN = mean, na.rm = TRUE)
  print(avg_scores)

  plot(window.size, avg_scores)
  lines(window.size, avg_scores, lwd = 2)
}

#' Reads Hi-C Contact Data from File
#' 
#' @param file The pathname of a normalize Hi-C contact matrix file
#' stored as a whitespace-delimited file.  See below for details.
#' Also a gzip-compressed file can be used.
#'
#' @param chr,binSize If the file contains a count matrix without bin
#' annotation, the latter is created from these parameters.
#'
#' @param \ldots Arguments passed to [utils::read.table()] as-is.
#'
#' @param debug If `TRUE`, debug output is produced.
#' 
#' @return A list with elements \code{bins} (an N-by-4 data.frame) and
#' \code{counts} (N-by-N matrix).
#'
#' @section Format of HiC contact-matrix file:
#' The contact-matrix file should be a whitespace-delimited text file with
#' neither row names nor column names.  The content should be a N-by-(3+N)
#' table where the first three columns correspond to `chr` (string),
#' `from.coord` (integer position), and `to.coord` (integer position).
#' These column defines the genomic location of the N Hi-C bins (in order).
#' The last N columns should contain normalized contact counts (float) such
#' that element (r,3+c) in this table corresponds to count (r,c) in the
#' normalized contact matrix.
#'
#' If an N-by-(4+N) table, then the first column is assumed to contain an
#' `id` (integer), and everything else as above.
#'
#' Example:
#' \preformatted{
#'   chr10       0   40000  0 0 0 0 ...
#'   chr10   40000   80000  0 0 0 0 ...
#'   chr10   80000  120000  0 0 0 0 ...
#'   chr10  120000  160000  0 0 0 0 ...
#'   ...
#' }
#'
#' @example incl/readHiC.R
#' 
#' @seealso [TopDom].
#'
#' @importFrom utils file_test read.table
#' @export
readHiC <- function(file, chr = NULL, binSize = NULL, ..., debug = getOption("TopDom.debug", FALSE)) {
  stopifnot(file_test("-f", file))
  stopifnot(is.logical(debug), length(debug) == 1L, !is.na(debug))

  if (debug) {
    mcat("#########################################################################")
    mcat("Step 0 : File Read")
    mcat("#########################################################################")
  }
  
  if (!is.null(chr)) {
    chr <- as.character(chr)
    stopifnot(is.character(chr), length(chr) == 1, !is.na(chr))
    binSize <- as.integer(binSize)
    stopifnot(is.integer(binSize), length(binSize) == 1, !is.na(binSize), binSize >= 1)

    args <- list(..., comment.char = "", na.strings = "", quote = "",
                      stringsAsFactors = FALSE)
    bins <- args$bins                        
    args$bins <- NULL
    args <- args[unique(names(args))]
    argsT <- c(list(file, header = FALSE, nrows = 1L), args)
    first <- do.call(read.table, args = argsT)
    if (debug) mcat("  -- reading ", length(first), "-by-", length(first), " count matrix")
    ## Assert that it's a count matrix
    is.numeric <- unlist(lapply(first, FUN = is.numeric), use.names = FALSE)
    stopifnot(all(is.numeric))
    
    if (!is.null(bins)) {
      if (any(bins < 1)) {
        stop("Argument 'bins' specifies non-positive bin indices")
      } else if (any(bins > length(first))) {
        stop(sprintf("Argument 'bins' specifies bin indices out of range [1,%d]", length(first)))
      }
      colClasses <- rep("NULL", times = length(first))
      colClasses[bins] <- "numeric"
    } else {
      colClasses <- rep("numeric", times = length(first))
    }
    argsT <- c(list(file, colClasses = colClasses, header = FALSE), args)
    matrix.data <- do.call(read.table, args = argsT)
    colnames(matrix.data) <- NULL

    if (!is.null(bins)) {
      matrix.data <- matrix.data[bins, , drop = FALSE]
    }

    ## N-by-N count matrix (from file content)
    matrix.data <- as.matrix(matrix.data)
    dimnames(matrix.data) <- NULL
    stopifnot(nrow(matrix.data) == ncol(matrix.data))
    n_bins <- length(first)

    from.coord <- seq(from = 0, by = binSize, length.out = n_bins)
    to.coord   <- seq(from = binSize, by = binSize, length.out = n_bins)
    if (is.null(bins)) {
      stopifnot(n_bins == nrow(matrix.data))
      id <- seq_len(n_bins)
    } else {
      n_bins <- length(bins)
      id <- bins
      from.coord <- from.coord[bins]
      to.coord   <- to.coord[bins]
    }
    
    ## Bin annotation from (chr, binSize)
    bins <- data.frame(
      id               = id,
      chr              = chr,
      from.coord       = from.coord,
      to.coord         = to.coord,
      stringsAsFactors = FALSE
    )
  } else {
    args <- list(..., stringsAsFactors = FALSE)
    args <- args[unique(names(args))]
    argsT <- c(list(file, header = FALSE), args)
    matdf <- do.call(read.table, args = argsT)
    n_bins <- nrow(matdf)
    if (ncol(matdf) - n_bins == 3) {
      colnames(matdf) <- c("chr", "from.coord", "to.coord")
    } else if (ncol(matdf) - n_bins == 4) {
      colnames(matdf) <- c("id", "chr", "from.coord", "to.coord")
    } else {
      stop("Unknown format of count-matrix file: ", sQuote(file))
    }

    ## Bin annotation (from file content)
    bins <- data.frame(
      id         = seq_len(n_bins),
      chr        = matdf[["chr"]],
      from.coord = matdf[["from.coord"]],
      to.coord   = matdf[["to.coord"]],
      stringsAsFactors = FALSE
    )

    ## N-by-N count matrix (from file content)
    matdf <- matdf[, (ncol(matdf) - n_bins + 1):ncol(matdf)]
    matrix.data <- as.matrix(matdf)
    rm(list = "matdf")
  }

  stopifnot(is.numeric(matrix.data),
            is.matrix(matrix.data),
            nrow(matrix.data) == ncol(matrix.data),
            nrow(matrix.data) == n_bins)

  if (debug) mcat("-- Done!")
  if (debug) mcat("Step 0 : Done!")

  structure(list(bins = bins, counts = matrix.data), class = "TopDomData")
}

library("TopDom")

stop_if_not <- TopDom:::stop_if_not
mcat <- TopDom:::mcat

## stop_if_not() - passing cases return invisibly (NULL)
res <- withVisible(stop_if_not(TRUE))
stopifnot(is.null(res$value), isFALSE(res$visible))
stop_if_not(TRUE, 1 == 1, !FALSE)

## stop_if_not() - failing cases signal an error, and the message contains
## a deparsed version of the failing expression
res <- tryCatch(stop_if_not(1 == 2), error = identity)
stopifnot(inherits(res, "error"))
stopifnot(grepl("1 == 2", conditionMessage(res)))

## Non-scalar / NA / non-logical arguments all fail
res <- tryCatch(stop_if_not(c(TRUE, TRUE)), error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch(stop_if_not(NA), error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch(stop_if_not(TRUE, FALSE), error = identity)
stopifnot(inherits(res, "error"))

## Long expressions are truncated with a trailing '....'
res <- tryCatch(
  stop_if_not(is.numeric("a really really really really really long expression here")),
  error = identity
)
stopifnot(inherits(res, "error"))

## mcat() writes to stderr and returns invisibly
err <- utils::capture.output(out <- withVisible(mcat("hello", " ", "world")), type = "message")
stopifnot(isFALSE(out$visible))
stopifnot(any(grepl("hello world", err)))

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")

pathname <- file.path(path, sprintf("nij.%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3)
data <- data[1:200]

fit <- TopDom(data, window.size = 5L)
domains <- subset(fit$domain, tag == "domain")
region <- domains[which.max(domains$size), ]

## (a) Subsetting a TopDomData object -----------------------------------------

## Relative margin (margin < 1)
sub_rel <- subsetByRegion(data, region = region, margin = 1/2)
print(sub_rel)
stopifnot(inherits(sub_rel, "TopDomData"))

## Absolute margin (margin >= 1)
sub_abs <- subsetByRegion(data, region = region, margin = 80e3)
stopifnot(inherits(sub_abs, "TopDomData"))

## Zero margin gives exactly the region
sub_zero <- subsetByRegion(data, region = region, margin = 0)
stopifnot(
  inherits(sub_zero, "TopDomData"),
  all(sub_zero$bins$from.coord >= region$from.coord),
  all(sub_zero$bins$to.coord <= region$to.coord),
  nrow(sub_zero$bins) <= nrow(sub_rel$bins)
)

## Negative margins are rejected
res <- tryCatch(subsetByRegion(data, region = region, margin = -1), error = identity)
stopifnot(inherits(res, "error"))

## 'region' must be a data.frame
res <- tryCatch(subsetByRegion(data, region = 42), error = identity)
stopifnot(inherits(res, "error"))

## (b) Subsetting a TopDom object -------------------------------------------

sub_td <- subsetByRegion(fit, region = region, margin = 1/2)
str(sub_td$domain)
stopifnot(
  inherits(sub_td, "TopDom"),
  is.data.frame(sub_td$domain),
  nrow(sub_td$domain) >= 1L
)

sub_td0 <- subsetByRegion(fit, region = region, margin = 0)
stopifnot(inherits(sub_td0, "TopDom"), nrow(sub_td0$domain) >= 1L)

sub_td_abs <- subsetByRegion(fit, region = region, margin = 200e3)
stopifnot(nrow(sub_td_abs$domain) >= nrow(sub_td0$domain))

res <- tryCatch(subsetByRegion(fit, region = "x"), error = identity)
stopifnot(inherits(res, "error"))

res <- tryCatch(subsetByRegion(fit, region = region, margin = -0.1), error = identity)
stopifnot(inherits(res, "error"))

library("TopDom")

path <- system.file("exdata", package = "TopDom", mustWork = TRUE)
chr <- Sys.getenv("R_TOPDOM_TESTS_CHROMOSOME", "chr19")

## Original count data (subset to keep the test fast)
pathname <- file.path(path, sprintf("nij.%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3)
data <- data[1:200]
str(data)

## Find topological domains
fit <- TopDom(data, window.size = 5L)

## Counts per region for the identified domains
domains <- subset(fit$domain, tag == "domain")
counts <- countsPerRegion(data, regions = domains)
str(counts)
stopifnot(
  is.numeric(counts),
  length(counts) == nrow(domains),
  all(counts >= 0),
  !anyNA(counts)
)

## A single region works too
counts1 <- countsPerRegion(data, regions = domains[1, ])
stopifnot(length(counts1) == 1L, counts1 == counts[1])

## Regions spanning more than one chromosome are not allowed
domains2 <- domains[1:2, ]
domains2$chr <- c(chr, "chrX")
res <- tryCatch(countsPerRegion(data, regions = domains2), error = identity)
stopifnot(inherits(res, "error"))

## 'regions' must be a data.frame
res <- tryCatch(countsPerRegion(data, regions = "not-a-data-frame"), error = identity)
stopifnot(inherits(res, "error"))

#' Subset a TopDomData Object by Region
#'
#' @param data A TopDomData object.
#'
#' @param region A TopDom domain (a data.frame).
#'
#' @param margin An non-negative numeric specifying the additional margin
#'               extracted around the domain.
#'               If `margin < 1`, then the size of the margin is relative
#'               to the size of the domain.
#'
#' @return A TopDomData object.
#'
#' @author Henrik Bengtsson.
#'
#' @export
subsetByRegion <- function(data, region, margin = 1/2) {
  UseMethod("subsetByRegion")
}


#' @export
subsetByRegion.TopDomData <- function(data, region, margin = 1/2) {
  stopifnot(is.data.frame(region))
  stopifnot(margin >= 0)
  
  if (margin < 1) {
    margin <- margin * (region$to.coord - region$from.coord)
  }
  
  idxs <- with(data$bins, which(chr == region$chr & from.coord >= region$from.coord - margin & to.coord <= region$to.coord + margin))
  
  data[idxs]
}


#' @export
subsetByRegion.TopDom <- function(data, region, margin = 1/2) {
  stopifnot(is.data.frame(region))
  stopifnot(margin >= 0)
  
  if (margin < 1) {
    margin <- margin * (region$to.coord - region$from.coord)
  }

  idxs <- with(data$domain, which(chr == region$chr & from.coord >= region$from.coord - margin & to.coord <= region$to.coord + margin))
  
  data[idxs, ]
}

#' Identify Topological Domains from a Hi-C Contact Matrix
#' 
#' @param data A TopDomData object, or the pathname to a normalized
#' Hi-C contact matrix file as read by [readHiC()], that specify N bins.
#' 
#' @param window.size The number of bins to extend (as a non-negative integer).
#' Recommended range is in \{5, ..., 20\}.
#' 
#' @param outFile (optional) The filename without extension of the three
#' result files optionally produced. See details below.
#' 
#' @param statFilter (logical) Specifies whether non-significant
#' topological-domain boundaries should be dropped or not.
#' 
#' @param ... Additional arguments passed to [readHiC()].
#' 
#' @param debug If `TRUE`, debug output is produced.
#'
#' @return A named list of class `TopDom` with data.frame elements
#' `binSignal`, `domain`, and `bed`.
#' * The `binSignal` data frame (N-by-7) holds mean contact frequency,
#'   local extreme, and p-value for every bin. The first four columns
#'   represent basic bin information given by matrix file, such as
#'   bin id (`id`), chromosome(`chr`), start coordinate (`from.coord`),
#'   and end coordinate (`to.coord`) for each bin.
#'   The last three columns (`local.ext`, `mean.cf`, and `p-value`) represent
#'   computed values by the TopDom algorithm.
#'   The columns are:
#'   - `id`: Bin ID
#'   - `chr`: Chromosome
#'   - `from.coord`: Start coordinate of bin
#'   - `to.coord`: End coordinate of bin
#'   - `local.ext`:
#'      + `-1`: Local minima.
#'      + `-0.5`: Gap region.
#'      + `0`: General bin.
#'      + `1`: Local maxima.
#'   - `mean.cf`: Average of contact frequencies between lower and upper
#'     regions for bin _i = 1,2,...,N_.
#'   - `p-value`: Computed p-value by Wilcox rank sum test.
#'     See Shin et al. (2016) for more details.
#' 
#' * The `domain` data frame (D-by-7):
#'   Every bin is categorized by basic building block, such as gap, domain,
#'   or boundary.
#'   Each row indicates a basic building block.
#'   The first five columns include the basic information about the block,
#'   'tag' column indicates the class of the building block.
#'   - `id`: Identifier of block
#'   - `chr`: Chromosome
#'   - `from.id`: Start bin index of the block
#'   - `from.coord`: Start coordinate of the block
#'   - `to.id`: End bin index of the block
#'   - `to.coord`: End coordinate of the block
#'   - `tag`: Categorized name of the block. Three possible blocks exists:
#'     + `gap`
#'     + `domain`
#'     + `boundary`
#'   - `size`: size of the block
#'
#' * The `bed` data frame (D-by-4) is a representation of the `domain`
#'   data frame in the
#'   [BED file format](https://genome.ucsc.edu/FAQ/FAQformat.html#format1).
#'   It has four columns:
#'   - `chrom`: The name of the chromosome.
#'   - `chromStart`: The starting position of the feature in the chromosome.
#'      The first base in a chromosome is numbered 0.
#'   - `chromEnd`: The ending position of the feature in the chromosome.
#'      The `chromEnd` base is _not_ included in the feature. For example,
#'      the first 100 bases of a chromosome are defined as `chromStart=0`,
#'      `chromEnd=100`, and span the bases numbered 0-99.
#'   - `name`: Defines the name of the BED line. This label is displayed to
#'      the left of the BED line in the
#'      [UCSC Genome Browser](https://genome.ucsc.edu/cgi-bin/hgGateway)
#'      window when the track is open to full display mode or directly to
#'      the left of the item in pack mode.
#' 
#' If argument `outFile` is non-`NULL`, then the three elements (`binSignal`,
#' `domain`, and `bed`) returned are also written to tab-delimited files
#' with file names \file{<outFile>.binSignal}, \file{<outFile>.domain}, and
#' \file{<outFile>.bed}, respectively.  None of the files have row names,
#' and all but the BED file have column names.
#'
#' @section Windows size:
#' The `window.size` parameter is by design the only tuning parameter in the
#' TopDom method and affects the amount of smoothing applied when calculating
#' the TopDom bin signals.  The binning window extends symmetrically downstream
#' and upstream from the bin such that the bin signal is the average
#' `window.size^2` contact frequencies.
#' For details, see Equation (1) and Figure 1 in Shin et al. (2016).
#' Typically, the number of identified TDs decreases while their average
#' lengths increase as this window-size parameter increases (Figure 2).
#' The default is `window.size = 5` (bins), which is motivated as:
#' "Considering the previously reported minimum TD size (approx. 200 kb)
#' (Dixon et al., 2012) and our bin size of 40 kb, _w_\[indow.size\] = 5 is a
#' reasonable setting" (Shin et al., 2016).
#'
#' @example incl/TopDom.R
#' 
#' @author Hanjun Shin, Harris Lazaris, and Gangqing Hu.
#' \R package, help, and code refactoring by Henrik Bengtsson.
#'
#' @references
#'
#' * Shin et al.,
#'   TopDom: an efficient and deterministic method for identifying
#'   topological domains in genomes,
#'   _Nucleic Acids Research_, 44(7): e70, April 2016.
#'   DOI: \doi{10.1093/nar/gkv1505},
#'   PMCID: PMC4838359,
#'   PMID: 26704975
#'
#' * Shin et al., \R script \file{TopDom_v0.0.2.R}, 2017 (originally from
#'   \code{http://zhoulab.usc.edu/TopDom/};
#'   later available on \url{https://github.com/jasminezhoulab/TopDom})
#'
#' * Shin et al., TopDom Manual, 2016-07-08 (original from
#'   \code{http://zhoulab.usc.edu/TopDom/TopDom\%20Manual_v0.0.2.pdf};
#'   later available on \url{https://github.com/jasminezhoulab/TopDom})
#'
#' * Hanjun Shin, Understanding the 3D genome organization in topological
#'   domain level, Doctor of Philosophy Dissertation,
#'   University of Southern California, March 2017.
#'
#' * Dixon JR, Selvaraj S, Yue F, Kim A, et al. Topological domains in
#'   mammalian genomes identified by analysis of chromatin interactions.
#'   _Nature_; 485(7398):376-80, April 2012.
#'   DOI: \doi{10.1038/nature11082},
#'   PMCID: PMC3356448,
#'   PMID: 22495300.
#'
#' @importFrom utils read.table write.table
#' @export
TopDom <- function(data, window.size, outFile = NULL, statFilter = TRUE, ..., debug = getOption("TopDom.debug", FALSE)) {
  window.size <- as.integer(window.size)
  stopifnot(is.numeric(window.size),
            length(window.size) == 1,
            !is.na(window.size),
            window.size >= 0)
  stopifnot(is.logical(debug), length(debug) == 1L, !is.na(debug))

  if (is.character(data)) data <- readHiC(data, ...)
  stopifnot(inherits(data, "TopDomData"))

  bins <- data$bins
  matrix.data <- data$counts
  n_bins <- nrow(bins)

  ## The matrix-scaling part of the statistical-filtering step (statFilter =
  ## TRUE) addresses elements of the N-by-N contact matrix by their linear
  ## index, up to N * N.  For a large enough N, that product overflows R's
  ## integer range, after which seq() fails with an obscure "'to' must be a
  ## finite number" error preceded by a "NAs produced by integer overflow"
  ## warning.  Fail early with an explanation instead.
  ## See <https://github.com/HenrikBengtsson/TopDom/issues/10>.
  if (statFilter && n_bins > sqrt(.Machine$integer.max)) {
    stop(sprintf("The statistical-filtering step of TopDom() (statFilter = TRUE) does not support contact matrices with more than %d bins; here the matrix has %d bins, for which the internal index arithmetic (n_bins * n_bins) would overflow R's integer range. Workarounds: (i) call TopDom() with statFilter = FALSE, or (ii) use a larger bin size so that the chromosome is represented by fewer bins", as.integer(sqrt(.Machine$integer.max)), n_bins))
  }

  mean.cf <- rep(0, times = n_bins)
  pvalue <- rep(1.0, times = n_bins)

  ## Gap region (== -0.5) by default
  local.ext <- rep(-0.5, times = n_bins)  ## gap region

  if (debug) {
    mcat("#########################################################################")
    mcat("Step 1 : Generating binSignals by computing bin-level contact frequencies")
    mcat("#########################################################################")
  }
  ptm <- proc.time()
  for (i in seq_len(n_bins)) {
    diamond <- Get.Diamond.Matrix(mat.data = matrix.data, i = i, size = window.size)
    mean.cf[i] <- mean(diamond)
  }
  stop_if_not(length(mean.cf) == n_bins)

  eltm <- proc.time() - ptm
  if (debug) {
    mcat("Step 1 Running Time : ", eltm[3])
    mcat("Step 1 : Done!")

    mcat("#########################################################################")
    mcat("Step 2 : Detect TD boundaries based on binSignals")
    mcat("#########################################################################")
  }

  ptm <- proc.time()
  # gap.idx <- Which.Gap.Region(matrix.data = matrix.data)
  # gap.idx <- Which.Gap.Region2(mean.cf)
  gap.idx <- Which.Gap.Region2(matrix.data = matrix.data, w = window.size)

  proc.regions <- Which.process.region(rmv.idx = gap.idx, n_bins = n_bins, min.size = 3L)
  stop_if_not(!anyNA(proc.regions[["start"]]), all(proc.regions[["start"]] >= 1), all(proc.regions[["start"]] <= n_bins))
  stop_if_not(!anyNA(proc.regions[["end"]]), all(proc.regions[["end"]] >= 1), all(proc.regions[["end"]] <= n_bins))

  # mcat(proc.regions)

  for (i in seq_len(nrow(proc.regions))) {
    start <- proc.regions[i, "start"]
    end   <- proc.regions[i,   "end"]
    if (debug) mcat("Process Region #", i, " from ", start, " to ", end)
    idxs <- start:end
    local.ext[idxs] <- Detect.Local.Extreme(x = mean.cf[idxs])  ## assigns values (-1,0,+1)
  }
  stop_if_not(!anyNA(local.ext), length(local.ext) == n_bins, all(local.ext %in% c(-0.5, -1, 0, +1)))
  ## 'idxs' is only created if there is at least one region to process
  if (exists("idxs", inherits = FALSE)) rm(list = "idxs")

  eltm <- proc.time() - ptm
  if (debug) {
    mcat("Step 2 Running Time : ", eltm[3])
    mcat("Step 2 : Done!")
  }
  
  if (statFilter) {
    if (debug) {
      mcat("#########################################################################")
      mcat("Step 3 : Statistical Filtering of false positive TD boundaries")
      mcat("#########################################################################")
    }

    ptm <- proc.time()
    if (debug) mcat("-- Matrix Scaling....")
    scale.matrix.data <- matrix.data
    scaled.idxs <- integer(0L)
    for (i in seq_len(2 * window.size)) {
      # diag(scale.matrix.data[, i:n_bins]) <- scale(diag(matrix.data[, i:n_bins]))
      idxs <- seq(from = 1 + (n_bins * i), to = n_bins * n_bins, by = 1 + n_bins)
      scale.matrix.data[idxs] <- scale(matrix.data[idxs])
      scaled.idxs <- c(scaled.idxs, idxs)
    }
    ## 'idxs' is only created if 2 * window.size >= 1
    if (exists("idxs", inherits = FALSE)) rm(list = "idxs")
    rm(list = "matrix.data")

    ## Standardizing an all-constant (typically all-zero) near-diagonal band
    ## produces non-finite values, because such a band has zero variance.
    ## These NaNs later propagate into Get.Pvalue() -> wilcox.test(), which
    ## then fails with an obscure "not enough (finite) 'x' observations"
    ## error.  Detect this here and explain the actual problem instead.
    ## See <https://github.com/HenrikBengtsson/TopDom/issues/17>.
    if (length(scaled.idxs) > 0L && any(!is.finite(scale.matrix.data[scaled.idxs]))) {
      stop("The statistical-filtering step of TopDom() (statFilter = TRUE) failed because standardizing the near-diagonal bands of the contact matrix produced non-finite values. This happens when one or more of these bands has zero variance (e.g. all contact counts are zero) or contains missing values, which is typically caused by very sparse input data (too few read pairs and/or too small a bin size). Possible workarounds: (i) call TopDom() with statFilter = FALSE, (ii) use a larger bin size, or (iii) drop empty / very-low-count bins before calling TopDom()")
    }
    rm(list = "scaled.idxs")
    
    if (debug) mcat("-- Compute p-values by Wilcox Ranksum Test")
    for (i in seq_len(nrow(proc.regions))) {
      start <- proc.regions[i, "start"]
      end <- proc.regions[i, "end"]

      if (debug) mcat("Process Region #", i, " from ", start, " to ", end)

      idxs <- start:end
      pvalue[idxs] <- Get.Pvalue(matrix.data = scale.matrix.data[idxs, idxs], size = window.size, scale = 1.0)
    }
    ## 'idxs' is only created if there is at least one region to process
    if (exists("idxs", inherits = FALSE)) rm(list = "idxs")
    stop_if_not(length(pvalue) == n_bins, !anyNA(pvalue))
    if (debug) mcat("-- Done!")

    if (debug) mcat("-- Filtering False Positives")
    ## NOTE: The below duplication is left on purpose until we fully
    ##       understand why it is there in the first place, cf.
    ##       https://github.com/HenrikBengtsson/TopDom/issues/3
#    local.ext[((local.ext == -1.0) | (local.ext == -1.0)) & (pvalue < 0.05)] <- -2.0
#    local.ext[local.ext == -1.0] <-  0.0  ## general bin
#    local.ext[local.ext == -2.0] <- -1.0  ## local minima
    ## "Finally, we filter out local minima with P-values larger than 0.05. [...]"
    ##  (Page 4 in Shin et al. 2016)
    local.ext[local.ext == -1.0 & pvalue >= 0.05] <- 0.0  ## drop non-significant local minima
    
    stop_if_not(!anyNA(local.ext), length(local.ext) == n_bins, all(local.ext %in% c(-0.5, -1, 0, +1)))
    
    if (debug) mcat("-- Done!")

    eltm <- proc.time() - ptm
    if (debug) mcat("Step 3 Running Time : ", eltm[3])
    if (debug) mcat("Step 3 : Done!")
  } else {
    rm(list = "matrix.data")
    pvalue <- 0
  }

  if (debug) {
    mcat("#########################################################################")
    mcat("Step 4 : Convert bins to domains (internal step)")
    mcat("#########################################################################")
  }
  
  domains <- Convert.Bin.To.Domain.TMP(
    bins = bins,
    signal.idx = which(local.ext == -1.0),  ## local minima
    gap.idx = which(local.ext == -0.5),     ## gap region
    pvalues = pvalue,
    pvalue.cut = 0.05
  )

  bins <- cbind(
    bins,
    local.ext = local.ext,
    mean.cf = mean.cf,
    pvalue = pvalue
  )

  bedform <- domains[, c("chr", "from.coord", "to.coord", "tag")]
  colnames(bedform) <- c("chrom", "chromStart", "chromEnd", "name")

  if (!is.null(outFile)) {
    if (debug) {
      mcat("#########################################################################")
      mcat("Writing Files")
      mcat("#########################################################################")
    }

    outBinSignal <- paste0(outFile, ".binSignal")
    if (debug) mcat("binSignal File : ", outBinSignal)
    write.table(bins, file = outBinSignal, quote = FALSE, row.names = FALSE, col.names = TRUE, sep = "\t")

    outDomain <- paste0(outFile, ".domain")
    if (debug) mcat("Domain File : ", outDomain)
    write.table(domains, file = outDomain, quote = FALSE, row.names = FALSE, col.names = TRUE, sep = "\t")

    outBed <- paste0(outFile, ".bed")
    if (debug) mcat("Bed File : ", outBed)
    write.table(bedform, file = outBed, quote = FALSE, row.names = FALSE, col.names = FALSE, sep = "\t")
  }

  if (debug) mcat("Done!")

  if (debug) mcat("Job Complete!")
  
  res <- structure(
    list(binSignal = bins, domain = domains, bed = bedform),
    class = "TopDom"
  )
  attr(res, "parameters") <- list(
    window.size = window.size,
    statFilter = statFilter
  )
  
  res
}

# @fn Get.Diamond.Matrix
#
# @param mat.data N-by-N numeric matrix, where each element indicate contact frequency
#
# @param i (integer) a bin index
#
# @param size (integer) the window size to expand from bin `i`
#
# @return A subset of the `mat.data` matrix.  If `i` == N, then a missing value is returned.
Get.Diamond.Matrix <- function(mat.data, i, size) {
  n_bins <- nrow(mat.data)
  if (i == n_bins) {
    na <- NA_real_
    storage.mode(na) <- storage.mode(mat.data)
    return(na)
  }

  lowerbound <- max(1, i - size + 1)
  upperbound <- min(i + size, n_bins)

  mat.data[lowerbound:i, (i + 1):upperbound]
}

# @fn Which.process.region
# @param rmv.idx : vector of idx, remove index vector
# @param n_bins : total number of bins
# @param min.size : (integer) minimum size of bins
# @retrun : data.frame of proc.regions
Which.process.region <- function(rmv.idx, n_bins, min.size = 3L) {
  gap.idx <- rmv.idx

  proc.regions <- data.frame(start = numeric(0), end = numeric(0), stringsAsFactors = FALSE)
  proc.set <- setdiff(seq_len(n_bins), gap.idx)
  n_proc.set <- length(proc.set)

  i <- 1
  while (i < n_proc.set) {
    start <- proc.set[i]
    j <- i + 1

    while (j <= n_proc.set) {
      if (proc.set[j] - proc.set[j - 1] <= 1) {
        j <- j + 1
      } else {
        proc.regions <- rbind(proc.regions, c(start = start, end = proc.set[j - 1]))
        i <- j
        break
      }
    }

    if (j >= n_proc.set) {
      proc.regions <- rbind(proc.regions, c(start = start, end = proc.set[j - 1]))
      break
    }
  }

  colnames(proc.regions) <- c("start", "end")
  proc.regions <- proc.regions[abs(proc.regions[, "end"] - proc.regions[, "start"]) >= min.size, ]

  proc.regions
}

# @fn Which.Gap.Region
# @breif version 0.0.1 used
# @param matrix.data : n by n matrix
# @return gap index
Which.Gap.Region <- function(matrix.data) {
  n_bins <- nrow(matrix.data)
  gap <- rep(0, times = n_bins)

  i <- 1
  while (i < n_bins) {
    j <- i + 1
    while (j <= n_bins) {
      idxs <- i:j
      if (sum(matrix.data[idxs, idxs]) == 0) {
        gap[idxs] <- -0.5
        j <- j + 1
        # if (j-i > 1) gap[idxs] <- -0.5
        # j <- j+1
      } else {
        break
      }
    }

    i <- j
  }

  idx <- which(gap == -0.5)
  idx
}

# @fn Which.Gap.Region3
# @param matrix.data : n by n matrix
# @return gap index
Which.Gap.Region3 <- function(mean.cf) {
  n_bins <- length(mean.cf)
  gapidx <- which(mean.cf == 0)

  gapidx
}

# @fn Which.Gap.Region2
# @breif version 0.0.2 used
# @param matrix.data : n by n matrix
# @return gap index
Which.Gap.Region2 <- function(matrix.data, w) {
  n_bins <- nrow(matrix.data)
  gap <- rep(0, times = n_bins)

  for (i in seq_len(n_bins)) {
    if (sum(matrix.data[i, max(1, i - w):min(i + w, n_bins)]) == 0) gap[i] <- -0.5
  }

  idx <- which(gap == -0.5)
  idx
}

# @fn Detect.Local.Extreme
# @param x : original signal to find local minima
# @return vector of local extremes, -1 if the index is local minimum, +1 if the index is local maxima, 0 otherwise.
Detect.Local.Extreme <- function(x) {
  n_bins <- length(x)
  ret <- rep(0, times = n_bins)  ## general bin (default)
  x[is.na(x)] <- 0               ## general bin

  if (n_bins <= 3) {
    ret[which.min(x)] <- -1      ## local minima
    ret[which.max(x)] <- +1      ## local maxima
    return(ret)
  }
  
  # Norm##################################################3
  new.point <- Data.Norm(x = seq_len(n_bins), y = x)
  x <- new.point$y
  ##################################################
  cp <- Change.Point(x = seq_len(n_bins), y = x)
  cp <- cp$cp

  ncp <- length(cp)
  if (ncp <= 2) return(ret)
  if (ncp == n_bins) return(ret)
  
  for (i in 2:(ncp-1)) {
    cp_b <- cp[i]
    cp_c <- cp[i-1]

    x_a <- x[cp_b-1]
    x_b <- x[cp_b]
    x_c <- x[cp_b+1]
    
    if (x_b >= x_a && x_b >= x_c) {
      ret[cp_b] <- +1  ## local maxima
    } else if (x_b < x_a && x_b < x_c) {
      ret[cp_b] <- -1  ## local minima
    }

    min.val <- min(x[cp_c], x_b)
    max.val <- max(x[cp_c], x_b)

    x_r <- x[cp_c:cp_b]
    if (min(x_r) < min.val) ret[cp_c-1 + which.min(x_r)] <- -1  ## local minima
    if (max(x_r) > max.val) ret[cp_c-1 + which.max(x_r)] <- +1  ## local maxima
  }

  ret
}

# @fn Data.Norm
# @param x : x axis vector
# @param x : y axis vector
# @return list of normalized x and y
Data.Norm <- function(x, y) {
  ret.x <- rep(0, times = length(x))
  ret.y <- rep(0, times = length(y))

  ret.x[1] <- x[1]
  ret.y[1] <- y[1]

  diff.x <- diff(x)
  diff.y <- diff(y)

  scale.x <- 1 / mean(abs(diff(x)))
  scale.y <- 1 / mean(abs(diff(y)))

  # mcat(scale.x)
  # mcat(scale.y)

  diff.x <- diff.x * scale.x
  diff.y <- diff.y * scale.y
  
  for (i in 2:length(x)) {
    ret.x[i] <- ret.x[i - 1] + diff.x[i - 1]
    ret.y[i] <- ret.y[i - 1] + diff.y[i - 1]
  }

  list(x = ret.x, y = ret.y)
}

# @fn Change.Point
# @param x : x axis vector
# @param x : y axis vector
# @return change point index in x vector,
# Note that the first and the last point will be always change point
Change.Point <- function(x, y) {
  if (length(x) != length(y)) {
    mcat("ERROR : The length of x and y should be the same")
    return(0)
  }

  n_bins <- length(x)
  Fv <- rep(NA_real_, times = n_bins)
  Ev <- rep(NA_real_, times = n_bins)
  cp <- 1

  i <- 1
  Fv[1] <- 0
  while (i < n_bins) {
    j <- i + 1
    Fv[j] <- sqrt((x[j] - x[i]) ^ 2 + (y[j] - y[i]) ^ 2)

    while (j < n_bins) {
      j <- j + 1
      k <- (i + 1):(j - 1)
      dx_ji <- x[j] - x[i]
      dy_ji <- y[j] - y[i]
      A <- sqrt(dx_ji^2 + dy_ji^2)
      Ev_j <- sum(abs(dy_ji*x[k] - dx_ji*y[k] - x[i]*y[j] + x[j]*y[i])) / A
      Ev[j] <-     Ev_j
      Fv[j] <- A - Ev_j

      #################################################
      # Not Original Code
      if (is.na(Fv[j]) || is.na(Fv[j - 1])) {
        j <- j - 1
        cp <- c(cp, j)
        break
      }
      #################################################### 3
      if (Fv[j] < Fv[j - 1]) {
        j <- j - 1
        cp <- c(cp, j)
        break
      }
    }
    i <- j
  }

  cp <- c(cp, n_bins)

  list(cp = cp, objF = Fv, errF = Ev)
}

# @fn Get.Pvalue
# @param matrix.data : matrix
# @param size : size to extend
# @param scale : scale parameter if necessary. deprecated parameter
# @return computed p-value vector
#' @importFrom stats wilcox.test
Get.Pvalue <- function(matrix.data, size, scale = 1.0) {
  n_bins <- nrow(matrix.data)
  pvalue <- rep(1, times = n_bins)

  for (i in seq_len(n_bins - 1)) {
    dia <- as.vector(Get.Diamond.Matrix2(matrix.data, i = i, size = size))
    ups <- as.vector(Get.Upstream.Triangle(matrix.data, i = i, size = size))
    downs <- as.vector(Get.Downstream.Triangle(matrix.data, i = i, size = size))

    wil.test <- wilcox.test(x = dia * scale, y = c(ups, downs), alternative = "less", exact = FALSE)
    pvalue[i] <- wil.test$p.value

    # mcat(i, " = ", wil.test$p.value)
  }

  pvalue[is.na(pvalue)] <- 1
  pvalue
}

# @fn Get.Upstream.Triangle
# @param mat.data : matrix data
# @param i : bin index
# @param size : size of window to extend
# @return upstream triangle matrix
Get.Upstream.Triangle <- function(mat.data, i, size) {
  n_bins <- nrow(mat.data)

  lower <- max(1, i - size)
  idxs <- lower:i
  tmp.mat <- mat.data[idxs, idxs]
  tmp.mat[upper.tri(tmp.mat, diag = FALSE)]
}

# @fn Get.Downstream.Triangle
# @param mat.data : matrix data
# @param i : bin index
# @param size : size of window to extend
# @return downstream triangle matrix
Get.Downstream.Triangle <- function(mat.data, i, size) {
  n_bins <- nrow(mat.data)
  if (i == n_bins) {
    na <- NA_real_
    storage.mode(na) <- storage.mode(mat.data)
    return(na)
  }

  upperbound <- min(i + size, n_bins)
  idxs <- (i + 1):upperbound
  tmp.mat <- mat.data[idxs, idxs]
  tmp.mat[upper.tri(tmp.mat, diag = FALSE)]
}

# @fn Get.Diamond.Matrix2
# @param mat.data : matrix data
# @param i : bin index
# @param size : size of window to extend
# @return diamond matrix
Get.Diamond.Matrix2 <- function(mat.data, i, size) {
  n_bins <- nrow(mat.data)
  na <- NA_real_
  storage.mode(na) <- storage.mode(mat.data)
  new.mat <- matrix(rep(na, times = size * size), nrow = size, ncol = size)

  for (k in seq_len(size)) {
    if (i - (k - 1) >= 1 && i < n_bins) {
      lower <- min(i + 1, n_bins)
      upper <- min(i + size, n_bins)

      new.mat[size - (k - 1), seq_len(upper - lower + 1)] <- mat.data[i - (k - 1), lower:upper]
    }
  }

  new.mat
}

# @fn Convert.Bin.To.Domain
# @param bins : bin information
# @param signal.idx : signal index
# @param signal.idx : gap index
# @param pvalues : pvalue vector
# @param pvalue.cut : pvalue threshold
# @return dataframe storing domain information
Convert.Bin.To.Domain <- function(bins, signal.idx, gap.idx, pvalues = NULL, pvalue.cut = NULL) {
  n_bins <- nrow(bins)
  ret <- data.frame(chr = character(0), from.id = numeric(0), from.coord = numeric(0), to.id = numeric(0), to.coord = numeric(0), tag = character(0), size = numeric(0), stringsAsFactors = FALSE)
  levels(x = ret[, "tag"]) <- c("domain", "gap", "boundary")

  rmv.idx <- setdiff(seq_len(n_bins), gap.idx)
  proc.region <- Which.process.region(rmv.idx, n_bins = n_bins, min.size = 0L)
  from.coord <- bins[proc.region[, "start"], "from.coord"]
  n_procs <- nrow(proc.region)
  zeros <- double(length = n_procs)
  gap <- data.frame(
    chr = rep(bins[1, "chr"], times = n_procs),
    from.id = zeros,
    from.coord = from.coord,
    to.id = zeros,
    to.coord = zeros,
    tag = rep("gap", times = n_procs),
    size = zeros,
    stringsAsFactors = FALSE
  )

  rmv.idx <- union(signal.idx, gap.idx)
  proc.region <- Which.process.region(rmv.idx, n_bins = n_bins, min.size = 0L)
  n_procs <- nrow(proc.region)
  from.coord <- bins[proc.region[, "start"], "from.coord"]
  zeros <- double(length = n_procs)
  domain <- data.frame(
    chr = rep(bins[1, "chr"], times = n_procs),
    from.id = zeros,
    from.coord = from.coord,
    to.id = zeros,
    to.coord = zeros,
    tag = rep("domain", times = n_procs),
    size = zeros,
    stringsAsFactors = FALSE
  )

  rmv.idx <- setdiff(seq_len(n_bins), signal.idx)
  proc.region <- as.data.frame(Which.process.region(rmv.idx, n_bins = n_bins, min.size = 1L))
  n_procs <- nrow(proc.region)
  if (n_procs > 0) {
    from.coord <- bins[proc.region[, "start"] + 1, "from.coord"]
    zeros <- double(length = n_procs)
    boundary <- data.frame(
      chr = rep(bins[1, "chr"], times = n_procs),
      from.id = zeros,
      from.coord = from.coord,
      to.id = zeros,
      to.coord = zeros,
      tag = rep("boundary", times = n_procs),
      size = zeros,
      stringsAsFactors = FALSE
    )
    ret <- rbind(ret, boundary)
  }

  ret <- rbind(gap, domain)
  ret <- ret[order(ret[, 3]), ]

  ret[, "to.coord"] <- c(ret[2:nrow(ret), "from.coord"], bins[n_bins, "to.coord"])
  ret[, "from.id"] <- match(ret[, "from.coord"], table = bins[, "from.coord"])
  ret[, "to.id"] <- match(ret[, "to.coord"], table = bins[, "to.coord"])
  ret[, "size"] <- ret[, "to.coord"] - ret[, "from.coord"]

  if (!is.null(pvalues) && !is.null(pvalue.cut)) {
    for (i in seq_len(nrow(ret))) {
      if (ret[i, "tag"] == "domain") {
        domain.bins.idx <- ret[i, "from.id"]:ret[i, "to.id"]
        p.value.constr <- which(pvalues[domain.bins.idx] < pvalue.cut)

        if (length(domain.bins.idx) == length(p.value.constr)) ret[i, "tag"] <- "boundary"
      }
    }
  }

  new.bdr.set <- stack.bdr <- stack.bdr.empty <- data.frame(
    chr = character(0),
    from.id = numeric(0),
    from.coord = numeric(0),
    to.id = numeric(0),
    to.coord = numeric(0),
    tag = character(0),
    size = numeric(0),
    stringsAsFactors = FALSE
  )

  i <- 1L
  while (i <= nrow(ret)) {
    if (ret[i, "tag"] == "boundary") {
      stack.bdr <- rbind(stack.bdr, ret[i, ])
    } else if (nrow(stack.bdr) > 0) {
      new.bdr <- data.frame(
        chr        = bins[1, "chr"],
        from.id    = min(stack.bdr[, "from.id"]),
        from.coord = min(stack.bdr[, "from.coord"]),
        to.id      = max(stack.bdr[, "to.id"]),
        to.coord   = max(stack.bdr[, "to.coord"]),
        tag        = "boundary",
        size       = max(stack.bdr[, "to.coord"]) - min(stack.bdr[, "from.coord"]),
        stringsAsFactors = FALSE
      )
      new.bdr.set <- rbind(new.bdr.set, new.bdr)
      stack.bdr <- stack.bdr.empty
    }

    i <- i + 1L
  }
  rm(list = c("stack.bdr", "stack.bdr.empty"))
  

  ret <- rbind(ret[ret[, "tag"] != "boundary", ], new.bdr.set)
  ret <- ret[order(ret[, "to.coord"]), ]

  ret
}


# @fn Convert.Bin.To.Domain
# @param bins : bin information
# @param signal.idx : signal index
# @param signal.idx : gap index
# @param pvalues : pvalue vector
# @param pvalue.cut : pvalue threshold
# @return dataframe storing domain information
Convert.Bin.To.Domain.TMP <- function(bins, signal.idx, gap.idx, pvalues = NULL, pvalue.cut = NULL) {
  n_bins <- nrow(bins)
  ret <- data.frame(chr = character(0), from.id = numeric(0), from.coord = numeric(0), to.id = numeric(0), to.coord = numeric(0), tag = character(0), size = numeric(0), stringsAsFactors = FALSE)
  levels(x = ret[, "tag"]) <- c("domain", "gap", "boundary")

  rmv.idx <- setdiff(seq_len(n_bins), gap.idx)
  proc.region <- Which.process.region(rmv.idx, n_bins = n_bins, min.size = 0L)
  from.coord <- bins[proc.region[, "start"], "from.coord"]
  n_procs <- nrow(proc.region)
  zeros <- double(length = n_procs)
  gap <- data.frame(chr = rep(bins[1, "chr"], times = n_procs), from.id = zeros, from.coord = from.coord, to.id = zeros, to.coord = zeros, tag = rep("gap", times = n_procs), size = zeros, stringsAsFactors = FALSE)

  rmv.idx <- union(signal.idx, gap.idx)
  proc.region <- Which.process.region(rmv.idx, n_bins = n_bins, min.size = 0L)
  n_procs <- nrow(proc.region)
  from.coord <- bins[proc.region[, "start"], "from.coord"]
  zeros <- double(length = n_procs)
  domain <- data.frame(chr = rep(bins[1, "chr"], times = n_procs), from.id = zeros, from.coord = from.coord, to.id = zeros, to.coord = zeros, tag = rep("domain", times = n_procs), size = zeros, stringsAsFactors = FALSE)

  rmv.idx <- setdiff(seq_len(n_bins), signal.idx)
  proc.region <- as.data.frame(Which.process.region(rmv.idx, n_bins = n_bins, min.size = 1L))
  n_procs <- nrow(proc.region)
  if (n_procs > 0) {
    from.coord <- bins[proc.region[, "start"] + 1, "from.coord"]
    zeros <- double(length = n_procs)
    boundary <- data.frame(chr = rep(bins[1, "chr"], times = n_procs), from.id = zeros, from.coord = from.coord, to.id = zeros, to.coord = zeros, tag = rep("boundary", times = n_procs), size = zeros, stringsAsFactors = FALSE)
    ret <- rbind(ret, boundary)
  }

  if (nrow(domain) == 0L) {
    ret <- gap
  } else {
    ret <- rbind(gap, domain)
    ret <- ret[order(ret[, 3]), ]

    ## FIXME: Below code assumes nrow(ret) >= 2
    ret[, "to.coord"] <- c(ret[2:nrow(ret), "from.coord"], bins[n_bins, "to.coord"])
    ret[, "from.id"] <- match(ret[, "from.coord"], table = bins[, "from.coord"])
    ret[, "to.id"] <- match(ret[, "to.coord"], table = bins[, "to.coord"])
    ret[, "size"] <- ret[, "to.coord"] - ret[, "from.coord"]

    if (!is.null(pvalues) && !is.null(pvalue.cut)) {
      for (i in seq_len(nrow(ret))) {
        if (ret[i, "tag"] == "domain") {
          domain.bins.idx <- ret[i, "from.id"]:ret[i, "to.id"]
          p.value.constr <- which(pvalues[domain.bins.idx] < pvalue.cut)
  
          if (length(domain.bins.idx) == length(p.value.constr)) ret[i, "tag"] <- "boundary"
        }
      }
    }
  }

  ret
}



#' @importFrom utils str
#' @export
print.TopDom <- function(x, ...) {
  cat(sprintf("%s:\n", class(x)))
  cat("Parameters:\n")
  params <- attr(x, "parameters")
  if (length(params) > 0L) {
    cat(sprintf("- window.size: %d\n", params$window.size))
    cat(sprintf("- statFilter: %s\n", params$statFilter))
  } else {
    cat(" - N/A\n")
  }
  cat("binSignal:\n")
  str(x$binSignal)
  cat("domain:\n")
  str(x$domain)
  cat("bed:\n")
  str(x$bed)
}

#' @export
dim.TopDom <- function(x) {
  dim(x$domain)
}

#' @export
`[.TopDom` <- function(x, i, ...) {
  structure(list(
    binSignal = x$binSignal,
    domain    = x$domain[i, , drop = FALSE],
    bed       = x$bed[i, , drop = FALSE]
  ), class = "TopDom")
}

library("TopDom")

## Known versions return an environment with the legacy API ------------
api1 <- legacy("0.0.1")
api2 <- legacy("0.0.2")
stopifnot(is.environment(api1), is.environment(api2))
stopifnot(is.function(api1$TopDom), is.function(api2$TopDom))
stopifnot(is.function(api1$Detect.Local.Extreme))

## Unknown version -> error -------------------------------------------
res <- tryCatch(legacy("9.9.9"), error = identity)
stopifnot(inherits(res, "error"),
          grepl("Unknown TopDom legacy version", conditionMessage(res)))

#' Calculates Counts per Region in a TopDomData Object
#'
#' @param data A TopDomData object.
#'
#' @param regions TopDom regions (a data.frame), e.g. domains.
#'
#' @return A numeric vector of length `nrow(regions)`.
#'
#' @author Henrik Bengtsson.
#'
#' @export
countsPerRegion <- function(data, regions) {
  UseMethod("countsPerRegion")
}


#' @importFrom matrixStats colSums2
#' @export
countsPerRegion.TopDomData <- function(data, regions) {
  chr <- NULL  ## To please R CMD check
  
  stopifnot(is.data.frame(regions))
  uchr <- unique(regions$chr)
  stopifnot(length(uchr) == 1L)

  bins <- subset(data$bins, chr == uchr)
  counts <- data$counts
  
  nregions <- nrow(regions)
  total <- vector(storage.mode(counts), length = nregions)

  for (kk in seq_len(nregions)) {
    region <- regions[kk, ]
    idxs_kk <- with(bins, which(from.coord >= region$from.coord & to.coord <= region$to.coord))
    total[kk] <- sum(colSums2(counts, rows = idxs_kk, cols = idxs_kk, na.rm = TRUE), na.rm = TRUE)
  }

  total
}

#' Produce a Count Heatmap
#'
#' @param data A TopDomData object.
#'
#' @param transform A function applied to the counts prior to generating
#'        heatmap colors.
#'
#' @param colors A named list to control to color scale.
#'
#' @param \ldots Not used.
#'
#' @return A [ggplot2::ggplot] object.
#'
#' @author Henrik Bengtsson.
#'
#' @seealso See [TopDom] for an example.
#'
#' @export
ggCountHeatmap <- function(data, transform, colors, ...) UseMethod("ggCountHeatmap")

#' @importFrom reshape2 melt
#' @importFrom ggplot2 ggplot aes geom_raster coord_fixed theme_void scale_fill_gradient2
#' @export
ggCountHeatmap.TopDomData <- function(data, transform = function(x) log2(x + 1), colors = c(na = "white", mid = "blue", high = "yellow"), ...) {
  ## To please R CMD check
  x <- y <- counts <- NULL
  
  keepUpperTriangle <- function(x) {
    x[lower.tri(x)] <- NA
    x
  }

  cols <- eval(formals(ggCountHeatmap.TopDomData)$colors)
  keys <- names(cols)
  keys <- keys[is.na(colors[keys])]
  colors[keys] <- cols[keys]
    
  dd <- keepUpperTriangle(data$counts)
  dd <- melt(dd, varnames = c("x", "y"), na.rm = TRUE, value.name = "counts")

  ## Genomic coordinate system
  mid <- with(data$bins, (from.coord + to.coord) / 2)
  dd$x <- mid[dd$x]
  dd$y <- mid[dd$y]

  gg <- ggplot(dd)
  gg <- gg + aes(x, y, fill = transform(counts))
  gg <- gg + geom_raster(show.legend = FALSE)
  gg <- gg + coord_fixed()
  gg <- gg + theme_void()
  gg <- gg + scale_fill_gradient2(limits = c(0, NA), na.value = colors["na"],
                                  mid = colors["mid"], high = colors["high"])
  gg
}


#' Add a Topological Domain to a Count Heatmap
#'
#' @param td A single-row data.frame.
#'
#' @param dx,delta,vline Absolute distance to heatmap.
#'   If `dx = NULL` (default), then `dx = delta * w + vline` where `w` is
#'   the width of the domain.
#'
#' @param size,color The thickness and color of the domain line.
#'
#' @return A [ggplot2::geom_segment] object to be added to the count heatmap.
#'
#' @importFrom ggplot2 geom_segment
#' @export
ggDomain <- function(td, dx = NULL, delta = 0.04, vline = 0, size = 2.0, color = "#666666") {
  x0 <- td$from.coord
  x1 <- td$to.coord
  if (is.null(dx)) dx <- delta * (x1 - x0) + vline
  gg <- geom_segment(aes(x = x0+dx, y = x0-dx, xend = x1+dx, yend = x1-dx),
                     color = color, size = size)
  attr(gg, "gg_params") <- list(x0 = x0, x1 = x1, width = x1 - x0, dx = dx, delta = delta, vline = vline)
  gg		     
}


#' Add a Topological Domain Label to a Count Heatmap
#'
#' @param td A single-row data.frame.
#'
#' @param fmt The [base::sprintf]-format string taking (chromosome, start, stop) as
#'        (string, numeric, numeric) input.
#'
#' @param rot The amount of rotation in \[0,360\] of label.
#'
#' @param dx,vjust The vertical adjustment of the label (relative to rotation)
#'
#' @param cex The scale factor of the label.
#'
#' @return A [ggplot2::ggproto] object to be added to the count heatmap.
#'
#' @importFrom ggplot2 annotation_custom
#' @importFrom grid gpar textGrob
#' @export
ggDomainLabel <- function(td, fmt = "%s: %.2f - %.2f Mbp", rot = 45, dx = 0, vjust = 2.5, cex = 1.5) {
  chr <- td$chr
  x0 <- td$from.coord + dx
  x1 <- td$to.coord + dx
  label <- sprintf(fmt, chr, x0/1e6, x1/1e6)
  grob <- textGrob(label = label, rot = rot, hjust = 0.5, vjust = vjust, gp = gpar(cex = cex))
  annotation_custom(grob = grob, ymin = x0, ymax = x1, xmin = x0, xmax = x1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/countsPerRegion.R
\name{countsPerRegion}
\alias{countsPerRegion}
\title{Calculates Counts per Region in a TopDomData Object}
\usage{
countsPerRegion(data, regions)
}
\arguments{
\item{data}{A TopDomData object.}

\item{regions}{TopDom regions (a data.frame), e.g. domains.}
}
\value{
A numeric vector of length \code{nrow(regions)}.
}
\description{
Calculates Counts per Region in a TopDomData Object
}
\author{
Henrik Bengtsson.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TopDom.R
\name{TopDom}
\alias{TopDom}
\title{Identify Topological Domains from a Hi-C Contact Matrix}
\usage{
TopDom(
  data,
  window.size,
  outFile = NULL,
  statFilter = TRUE,
  ...,
  debug = getOption("TopDom.debug", FALSE)
)
}
\arguments{
\item{data}{A TopDomData object, or the pathname to a normalized
Hi-C contact matrix file as read by \code{\link[=readHiC]{readHiC()}}, that specify N bins.}

\item{window.size}{The number of bins to extend (as a non-negative integer).
Recommended range is in \{5, ..., 20\}.}

\item{outFile}{(optional) The filename without extension of the three
result files optionally produced. See details below.}

\item{statFilter}{(logical) Specifies whether non-significant
topological-domain boundaries should be dropped or not.}

\item{...}{Additional arguments passed to \code{\link[=readHiC]{readHiC()}}.}

\item{debug}{If \code{TRUE}, debug output is produced.}
}
\value{
A named list of class \code{TopDom} with data.frame elements
\code{binSignal}, \code{domain}, and \code{bed}.
\itemize{
\item The \code{binSignal} data frame (N-by-7) holds mean contact frequency,
local extreme, and p-value for every bin. The first four columns
represent basic bin information given by matrix file, such as
bin id (\code{id}), chromosome(\code{chr}), start coordinate (\code{from.coord}),
and end coordinate (\code{to.coord}) for each bin.
The last three columns (\code{local.ext}, \code{mean.cf}, and \code{p-value}) represent
computed values by the TopDom algorithm.
The columns are:
\itemize{
\item \code{id}: Bin ID
\item \code{chr}: Chromosome
\item \code{from.coord}: Start coordinate of bin
\item \code{to.coord}: End coordinate of bin
\item \code{local.ext}:
\itemize{
\item \code{-1}: Local minima.
\item \code{-0.5}: Gap region.
\item \code{0}: General bin.
\item \code{1}: Local maxima.
}
\item \code{mean.cf}: Average of contact frequencies between lower and upper
regions for bin \emph{i = 1,2,...,N}.
\item \code{p-value}: Computed p-value by Wilcox rank sum test.
See Shin et al. (2016) for more details.
}
\item The \code{domain} data frame (D-by-7):
Every bin is categorized by basic building block, such as gap, domain,
or boundary.
Each row indicates a basic building block.
The first five columns include the basic information about the block,
'tag' column indicates the class of the building block.
\itemize{
\item \code{id}: Identifier of block
\item \code{chr}: Chromosome
\item \code{from.id}: Start bin index of the block
\item \code{from.coord}: Start coordinate of the block
\item \code{to.id}: End bin index of the block
\item \code{to.coord}: End coordinate of the block
\item \code{tag}: Categorized name of the block. Three possible blocks exists:
\itemize{
\item \code{gap}
\item \code{domain}
\item \code{boundary}
}
\item \code{size}: size of the block
}
\item The \code{bed} data frame (D-by-4) is a representation of the \code{domain}
data frame in the
\href{https://genome.ucsc.edu/FAQ/FAQformat.html#format1}{BED file format}.
It has four columns:
\itemize{
\item \code{chrom}: The name of the chromosome.
\item \code{chromStart}: The starting position of the feature in the chromosome.
The first base in a chromosome is numbered 0.
\item \code{chromEnd}: The ending position of the feature in the chromosome.
The \code{chromEnd} base is \emph{not} included in the feature. For example,
the first 100 bases of a chromosome are defined as \code{chromStart=0},
\code{chromEnd=100}, and span the bases numbered 0-99.
\item \code{name}: Defines the name of the BED line. This label is displayed to
the left of the BED line in the
\href{https://genome.ucsc.edu/cgi-bin/hgGateway}{UCSC Genome Browser}
window when the track is open to full display mode or directly to
the left of the item in pack mode.
}
}

If argument \code{outFile} is non-\code{NULL}, then the three elements (\code{binSignal},
\code{domain}, and \code{bed}) returned are also written to tab-delimited files
with file names \file{<outFile>.binSignal}, \file{<outFile>.domain}, and
\file{<outFile>.bed}, respectively.  None of the files have row names,
and all but the BED file have column names.
}
\description{
Identify Topological Domains from a Hi-C Contact Matrix
}
\section{Windows size}{

The \code{window.size} parameter is by design the only tuning parameter in the
TopDom method and affects the amount of smoothing applied when calculating
the TopDom bin signals.  The binning window extends symmetrically downstream
and upstream from the bin such that the bin signal is the average
\code{window.size^2} contact frequencies.
For details, see Equation (1) and Figure 1 in Shin et al. (2016).
Typically, the number of identified TDs decreases while their average
lengths increase as this window-size parameter increases (Figure 2).
The default is \code{window.size = 5} (bins), which is motivated as:
"Considering the previously reported minimum TD size (approx. 200 kb)
(Dixon et al., 2012) and our bin size of 40 kb, \emph{w}[indow.size] = 5 is a
reasonable setting" (Shin et al., 2016).
}

\examples{
path <- system.file("exdata", package = "TopDom", mustWork = TRUE)

## Original count data (on a subset of the bins to speed up example)
chr <- "chr19"
pathname <- file.path(path, sprintf("nij.\%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3, bins = 1:500)
print(data)  ## a TopDomData object

## Find topological domains using the TopDom method
fit <- TopDom(data, window.size = 5L)
print(fit)  ## a TopDom object

## Display the largest domain
td <- subset(subset(fit$domain, tag == "domain"), size == max(size))
print(td) ## a data.frame

## Subset TopDomData object
data_s <- subsetByRegion(data, region = td, margin = 0.9999)
print(data_s)  ## a TopDomData object

vp <- grid::viewport(angle = -45, width = 0.7, y = 0.3)
gg <- ggCountHeatmap(data_s)
gg <- gg + ggDomain(td, color = "#cccc00") + ggDomainLabel(td)
print(gg, newpage = TRUE, vp = vp)

gg <- ggCountHeatmap(data_s, colors = list(mid = "white", high = "black"))
gg_td <- ggDomain(td, delta = 0.08)
dx <- attr(gg_td, "gg_params")$dx
gg <- gg + gg_td + ggDomainLabel(td, vjust = 2.5)
print(gg, newpage = TRUE, vp = vp)

## Subset TopDom object
fit_s <- subsetByRegion(fit, region = td, margin = 0.9999)
print(fit_s)  ## a TopDom object
for (kk in seq_len(nrow(fit_s$domain))) {
  gg <- gg + ggDomain(fit_s$domain[kk, ], dx = dx * (4 + kk \%\% 2), color = "red", size = 1)
}

print(gg, newpage = TRUE, vp = vp)


gg <- ggCountHeatmap(data_s)
gg_td <- ggDomain(td, delta = 0.08)
dx <- attr(gg_td, "gg_params")$dx
gg <- gg + gg_td + ggDomainLabel(td, vjust = 2.5)
fit_s <- subsetByRegion(fit, region = td, margin = 0.9999)
for (kk in seq_len(nrow(fit_s$domain))) {
  gg <- gg + ggDomain(fit_s$domain[kk, ], dx = dx * (4 + kk \%\% 2), color = "blue", size = 1)
}

print(gg, newpage = TRUE, vp = vp)
}
\references{
\itemize{
\item Shin et al.,
TopDom: an efficient and deterministic method for identifying
topological domains in genomes,
\emph{Nucleic Acids Research}, 44(7): e70, April 2016.
DOI: \doi{10.1093/nar/gkv1505},
PMCID: PMC4838359,
PMID: 26704975
\item Shin et al., \R script \file{TopDom_v0.0.2.R}, 2017 (originally from
\code{http://zhoulab.usc.edu/TopDom/};
later available on \url{https://github.com/jasminezhoulab/TopDom})
\item Shin et al., TopDom Manual, 2016-07-08 (original from
\code{http://zhoulab.usc.edu/TopDom/TopDom\%20Manual_v0.0.2.pdf};
later available on \url{https://github.com/jasminezhoulab/TopDom})
\item Hanjun Shin, Understanding the 3D genome organization in topological
domain level, Doctor of Philosophy Dissertation,
University of Southern California, March 2017.
\item Dixon JR, Selvaraj S, Yue F, Kim A, et al. Topological domains in
mammalian genomes identified by analysis of chromatin interactions.
\emph{Nature}; 485(7398):376-80, April 2012.
DOI: \doi{10.1038/nature11082},
PMCID: PMC3356448,
PMID: 22495300.
}
}
\author{
Hanjun Shin, Harris Lazaris, and Gangqing Hu.
\R package, help, and code refactoring by Henrik Bengtsson.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exdata.R
\docType{data}
\name{TopDom-data}
\alias{TopDom-data}
\title{Data for the TopDom Package}
\description{
The \file{exdata/} folder of this package provides a example data set
used in examples.  The data are also used to validate the \pkg{TopDom}
implementation toward the original TopDom scripts.
}
\section{Origin}{

The data herein contain a tiny subset of the HiC and TopDom data used
in the TopDom study (Shin et al., 2016).
More precisely, it contains:
\enumerate{
\item A TopDom file \file{mESC_5w_chr19.nij.HindIII.comb.40kb.domain}, which
is part of the \file{mESC_5w_domain.zip} file
(5,504 bytes; md5 ffb19996f681a4d35d5c9944f2c44343) from the
Supplementary Materials of Shin et al. (2016).
These data were downloaded from the
TopDom website (http://zhoulab.usc.edu/TopDom/ - now defunct).
\item A normalized HiC-count matrix file \file{nij.chr19.gz}, where the
non-compressed version is part of the \file{mESC.norm.tar.gz} file
(1,305,763,679 bytes; md5 2e79d0f57463b5b7c4bf86b187086d3c) made available
by \href{http://renlab.sdsc.edu/renlab_website/}{UCSD Ren Lab} originally
downloaded from \file{http://chromosome.sdsc.edu/mouse/hi-c/download.html}
(no longer available).
It is a tab-delimited file containing a 3250-by-3250 numeric matrix
non-negative decimal values. The underlying HiC sequence data is
available from
\href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE35156}{GSE35156}
on GEO and was published part of Dixon, et al. (2012).
}
}

\references{
\enumerate{
\item Dixon JR, Selvaraj S, Yue F, Kim A, et al. Topological domains in
mammalian genomes identified by analysis of chromatin interactions.
Nature 2012 Apr 11; 485(7398):376-80,
doi: \doi{10.1038/nature11082},
PMCID: PMC3356448,
PMID: 22495300.
\item Shin, et al., TopDom: an efficient and deterministic method for
identifying topological domains in genomes,
Nucleic Acids Res. 2016 Apr 20; 44(7): e70., 2016.
doi: \doi{10.1093/nar/gkv1505},
PMCID: PMC4838359,
PMID: 26704975.
}
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overlapScores.R
\name{overlapScores}
\alias{overlapScores}
\title{Calculates Overlap Scores Between Two Sets of Topological Domains}
\usage{
overlapScores(a, reference, debug = getOption("TopDom.debug", FALSE))
}
\arguments{
\item{a, reference}{Topological domain (TD) set \eqn{A} and TD reference
set \eqn{R} both in a format as returned by \code{\link[=TopDom]{TopDom()}}.}

\item{debug}{If \code{TRUE}, debug output is produced.}
}
\value{
Returns a named list of class \code{TopDomOverlapScores}, where the names
correspond to the chromosomes in domain reference set \eqn{R}.
Each of these chromosome elements contains a data.frame with fields:
\itemize{
\item \code{chromosome}  - \eqn{D_{R,c}} character strings
\item \code{best_score}  - \eqn{D_{R,c}} numerics in \eqn{[0,1]}
\item \code{best_length} - \eqn{D_{R,c}} positive integers
\item \code{best_set}    - list of \eqn{D_{R,c}} index vectors
}

where \eqn{D_{R,c}} is the number of TDs in reference set \eqn{R} on
chromosome \eqn{c}.  If a TD in reference \eqn{R} is not a \code{"domain"},
then the corresponding \code{best_score} and \code{best_length} values are
\code{NA_real_} and \code{NA_integer_}, respectively, while \code{best_set} is an empty
list.
}
\description{
Calculates Overlap Scores Between Two Sets of Topological Domains
}
\details{
The \emph{overlap score}, \eqn{overlap(A', r_i)}, represents how well a
\emph{consecutive} subset \eqn{A'} of topological domains (TDs) in \eqn{A}
overlap with topological domain \eqn{r_i} in reference set \eqn{R}.
For each reference TD \eqn{r_i}, the \emph{best match} \eqn{A'_{max}} is
identified, that is, the \eqn{A'} subset that maximize
\eqn{overlap(A', r_i)}.
For exact definitions, see Page 8 in Shin et al. (2016).

Note that the overlap score is an asymmetric score, which means that
\code{overlapScores(a, b) != overlapScores(b, a)}.
}
\section{Warning - This might differ not be the correct implementation}{

The original TopDom scripts do not provide an implementation for
calculating overlap scores.  Instead, the implementation of
\code{TopDom::overlapScores()} is based on the textual description of
overlap scores provided in Shin et al. (2016).  It is not known if this
is the exact same algorithm and implementation as the authors of the
TopDom article used.
}

\examples{
library(tibble)
path <- system.file("exdata", package = "TopDom", mustWork = TRUE)

## Original count data (on a subset of the bins to speed up example)
chr <- "chr19"
pathname <- file.path(path, sprintf("nij.\%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3, bins = 1:500)
print(data)

## Find topological domains using TopDom method for two window sizes
tds_5 <- TopDom(data, window.size = 5L)
tds_6 <- TopDom(data, window.size = 6L)

## Overlap scores (in both directions)
overlap_56 <- overlapScores(tds_6, reference = tds_5)
print(overlap_56)
print(as_tibble(overlap_56))

overlap_65 <- overlapScores(tds_5, reference = tds_6)
print(overlap_65)
print(as_tibble(overlap_65))

}
\references{
\itemize{
\item Shin et al.,
TopDom: an efficient and deterministic method for identifying
topological domains in genomes,
\emph{Nucleic Acids Research}, 44(7): e70, April 2016.
doi: \doi{10.1093/nar/gkv1505},
PMCID: PMC4838359,
PMID: 26704975
}
}
\seealso{
\link{TopDom}.
}
\author{
Henrik Bengtsson - based on the description in Shin et al. (2016).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggCountHeatmap.R
\name{ggCountHeatmap}
\alias{ggCountHeatmap}
\title{Produce a Count Heatmap}
\usage{
ggCountHeatmap(data, transform, colors, ...)
}
\arguments{
\item{data}{A TopDomData object.}

\item{transform}{A function applied to the counts prior to generating
heatmap colors.}

\item{colors}{A named list to control to color scale.}

\item{\ldots}{Not used.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2::ggplot} object.
}
\description{
Produce a Count Heatmap
}
\seealso{
See \link{TopDom} for an example.
}
\author{
Henrik Bengtsson.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legacy.R
\name{legacy}
\alias{legacy}
\title{Easy Access to the Original TopDom 0.0.1 and 0.0.2 Implementations}
\usage{
legacy(version = c("0.0.1", "0.0.2"))
}
\arguments{
\item{version}{A version string.}
}
\value{
An environment containing the legacy TopDom API.
}
\description{
Easy Access to the Original TopDom 0.0.1 and 0.0.2 Implementations
}
\examples{
TopDom::legacy("0.0.2")$TopDom
TopDom::legacy("0.0.1")$Detect.Local.Extreme

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subsetByRegion.R
\name{subsetByRegion}
\alias{subsetByRegion}
\title{Subset a TopDomData Object by Region}
\usage{
subsetByRegion(data, region, margin = 1/2)
}
\arguments{
\item{data}{A TopDomData object.}

\item{region}{A TopDom domain (a data.frame).}

\item{margin}{An non-negative numeric specifying the additional margin
extracted around the domain.
If \code{margin < 1}, then the size of the margin is relative
to the size of the domain.}
}
\value{
A TopDomData object.
}
\description{
Subset a TopDomData Object by Region
}
\author{
Henrik Bengtsson.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{stop_if_not}
\alias{stop_if_not}
\title{Asserts the Truth of R Expressions}
\usage{
stop_if_not(...)
}
\arguments{
\item{\dots}{Zero or more \R expressions to be asserted to be TRUE.}
}
\value{
Nothing.
}
\description{
Asserts the Truth of R Expressions
}
\details{
A bare bone, faster version of \link[base:stopifnot]{base::stopifnot}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggCountHeatmap.R
\name{ggDomain}
\alias{ggDomain}
\title{Add a Topological Domain to a Count Heatmap}
\usage{
ggDomain(td, dx = NULL, delta = 0.04, vline = 0, size = 2, color = "#666666")
}
\arguments{
\item{td}{A single-row data.frame.}

\item{dx, delta, vline}{Absolute distance to heatmap.
If \code{dx = NULL} (default), then \code{dx = delta * w + vline} where \code{w} is
the width of the domain.}

\item{size, color}{The thickness and color of the domain line.}
}
\value{
A \link[ggplot2:geom_segment]{ggplot2::geom_segment} object to be added to the count heatmap.
}
\description{
Add a Topological Domain to a Count Heatmap
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readHiC.R
\name{readHiC}
\alias{readHiC}
\title{Reads Hi-C Contact Data from File}
\usage{
readHiC(
  file,
  chr = NULL,
  binSize = NULL,
  ...,
  debug = getOption("TopDom.debug", FALSE)
)
}
\arguments{
\item{file}{The pathname of a normalize Hi-C contact matrix file
stored as a whitespace-delimited file.  See below for details.
Also a gzip-compressed file can be used.}

\item{chr, binSize}{If the file contains a count matrix without bin
annotation, the latter is created from these parameters.}

\item{debug}{If \code{TRUE}, debug output is produced.}

\item{\ldots}{Arguments passed to \code{\link[utils:read.table]{utils::read.table()}} as-is.}
}
\value{
A list with elements \code{bins} (an N-by-4 data.frame) and
\code{counts} (N-by-N matrix).
}
\description{
Reads Hi-C Contact Data from File
}
\section{Format of HiC contact-matrix file}{

The contact-matrix file should be a whitespace-delimited text file with
neither row names nor column names.  The content should be a N-by-(3+N)
table where the first three columns correspond to \code{chr} (string),
\code{from.coord} (integer position), and \code{to.coord} (integer position).
These column defines the genomic location of the N Hi-C bins (in order).
The last N columns should contain normalized contact counts (float) such
that element (r,3+c) in this table corresponds to count (r,c) in the
normalized contact matrix.

If an N-by-(4+N) table, then the first column is assumed to contain an
\code{id} (integer), and everything else as above.

Example:
\preformatted{
  chr10       0   40000  0 0 0 0 ...
  chr10   40000   80000  0 0 0 0 ...
  chr10   80000  120000  0 0 0 0 ...
  chr10  120000  160000  0 0 0 0 ...
  ...
}
}

\examples{
path <- system.file("exdata", package = "TopDom", mustWork = TRUE)

## Original count data
chr <- "chr19"
pathname <- file.path(path, sprintf("nij.\%s.gz", chr))
data <- readHiC(pathname, chr = chr, binSize = 40e3)
print(data)
str(data)
}
\seealso{
\link{TopDom}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggCountHeatmap.R
\name{ggDomainLabel}
\alias{ggDomainLabel}
\title{Add a Topological Domain Label to a Count Heatmap}
\usage{
ggDomainLabel(
  td,
  fmt = "\%s: \%.2f - \%.2f Mbp",
  rot = 45,
  dx = 0,
  vjust = 2.5,
  cex = 1.5
)
}
\arguments{
\item{td}{A single-row data.frame.}

\item{fmt}{The \link[base:sprintf]{base::sprintf}-format string taking (chromosome, start, stop) as
(string, numeric, numeric) input.}

\item{rot}{The amount of rotation in [0,360] of label.}

\item{dx, vjust}{The vertical adjustment of the label (relative to rotation)}

\item{cex}{The scale factor of the label.}
}
\value{
A \link[ggplot2:ggproto]{ggplot2::ggproto} object to be added to the count heatmap.
}
\description{
Add a Topological Domain Label to a Count Heatmap
}
